#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""handlers.py: Custom logging handlers."""

import logging
import geniac.cli.utils.errorcounter

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2021"


class LogMixin:
    """Add logger property and error/warning/info tracking"""

    def __init__(self, *args, **kwargs):
        self._error_flag = False
        super().__init__(*args, **kwargs)

    @property
    def logger(self):
        """Logging logger instance"""
        return logging.getLogger(".".join([__name__, self.__class__.__name__]))

    @property
    def error_flag(self):
        """Trace of error message"""
        return self._error_flag

    @error_flag.setter
    def error_flag(self, value: bool):
        """Set the error flag"""
        self._error_flag = value

    def error(self, *args, **kwargs):
        """Log error message and keep a trace of it"""
        self.error_flag = True
        return self.logger.error(*args, **kwargs)

    def info(self, *args, **kwargs):
        """Log info messages"""
        return self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        """Log warning messages"""
        return self.logger.warning(*args, **kwargs)

    def debug(self, *args, **kwargs):
        """Log debug messages"""
        return self.logger.debug(*args, **kwargs)

    def critical(self, *args, **kwargs):
        """Log critical messages"""
        return self.logger.critical(*args, **kwargs)

    def exception(self, *args, **kwargs):
        """Log exception messages"""
        return self.logger.exception(*args, **kwargs)


class ExitOnExceptionHandler(logging.StreamHandler):
    """Custom handler that exit if the log level is critical"""

    def emit(self, record):
        """Pass log messages on to its super and exit if the log level is critical"""
        super().emit(record)
        if record.levelno is logging.CRITICAL:
            raise SystemExit(-1)
        if record.levelno >= logging.ERROR:
            geniac.cli.utils.errorcounter.error_counter = geniac.cli.utils.errorcounter.error_counter + 1




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""install.py: Geniac install command"""

import logging
import re
import os
import subprocess
import sys
from gettext import gettext as _
from importlib.resources import files
from pathlib import Path

from geniac.cli.commands.init import GeniacInit
from geniac.cli.utils.base import CMAKE_OPTION_PREFIX, MethodRecord

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacInstall(GeniacInit):
    """Geniac configuration file generator"""

    # CMAKE FILES ANALYZED TO GET CACHE VARIABLES
    CMAKE_CACHE_FILES = [files("geniac").joinpath("cmake/stepSetVariables.cmake")]
    # https://regex101.com/r/bppizX/1
    CMAKE_CACHE_RE = re.compile(
        r"(?P<cmake_set>set\((?P<cmake_var>\w+)\s+(?P<cmake_values>("
        r"(?P<cmake_value_quote>[\"'])(?P<cmake_value>[\"':.\w/,\-(]*))"
        r"(?P=cmake_value_quote)\s+)*CACHE\s+(?P<cmake_type>\w+)\s+"
        r"(?P<cmake_docstring>(?P<quote>(?P<simple>')|(?P<double>\"))"
        r"(?(simple)[\":.\w\s/,\-()]+|[\':.\w\s/,\-()]+)(?P=quote))\s*\))"
    )

    def __init__(
        self,
        *args,
        project_path: str = None,
        install_path: str = None,
        no_pre_clean: bool = False,
        no_post_clean: bool = False,
        **kwargs,
    ):
        """Init flags specific to GCheck command"""

        super().__init__(
            *args,
            project_path=project_path,
            init_build=False,
            pre_clean=True,
            post_clean=not no_post_clean,
            **kwargs,
        )

        # install mode
        self.mode = kwargs['mode']

        # Should we clean build dir if it has been initialized
        self.pre_clean = not no_pre_clean
        self.install_path = (
            Path(install_path).resolve()
            if install_path
            else self.working_dir / "install"
        )
        # add cmake options in a specific scope
        self.cmake_options = self._get_cmake_options(**kwargs)
        self._is_sudo = None

    @property
    def is_sudo(self):
        """Should we launch make step in sudo mode ?"""
        if not self._is_sudo:
            self._is_sudo = self._check_sudo()
        return self._is_sudo

    def _get_cmake_options(self, mode: str = None, **kwargs) -> list:
        """
        Get cmake options in kwargs and merge them with the ones coming from a specific mode
        (if mode_option is not empty)

        Args:
            mode_option: install mode configured in geniac.install.modes section
            **kwargs:

        Returns:

        """
        cmake_options = [
            "".join(
                [
                    f"-D{option.removeprefix(CMAKE_OPTION_PREFIX)}",
                    f'={value}'.strip()
                    if not isinstance(value, bool) and value
                    else "=ON"
                    if value
                    else "=OFF",
                ]
            )
            for option, value in kwargs.items()
            if option.startswith(CMAKE_OPTION_PREFIX) and value
        ]
        if mode:
            mode_section = f"{self.config_section}.modes"
            cmake_mode_options = (
                self.default_config.get(mode_section, mode).strip().split("\n")
                if self.default_config.has_option(mode_section, mode)
                else []
            )
            cmake_options = [
                _ for _ in cmake_options if _ not in cmake_mode_options
            ] + cmake_mode_options
        return cmake_options

    def _check_sudo(self) -> bool:
        """Guess if the CMake/Make commands should be run in sudo mode"""

        is_sudo = False

        if os.environ.get('USER') == "root":
            return False

        cmake_sudo_options = (
            self.default_config.get(self.config_section, "cmakeSudoOptions")
            .strip()
            .split("\n")
        )
        cmake_fakeroot_options = (
            self.default_config.get(self.config_section, "cmakeFakeRootOptions")
            .strip()
            .split("\n")
        )
        for cmake_option in self.cmake_options:
            if cmake_option in cmake_sudo_options:
                is_sudo = True
            if cmake_option in cmake_fakeroot_options:
                return False
        return is_sudo

    @staticmethod
    def _get_cmake_cache_vars(content: str) -> list:
        """Get content of CMake CACHE variables from CMake file content"""
        return [
            match.groupdict()
            for match in GeniacInstall.CMAKE_CACHE_RE.finditer(content)
        ]

    @staticmethod
    def get_cmake_cache_vars():
        """Get cmake cache vars from cmake files found in the package"""
        cmake_cache_vars = []
        for cmake_file in GeniacInstall.CMAKE_CACHE_FILES:
            if cmake_file.is_file():
                cmake_cache_vars += GeniacInstall._get_cmake_cache_vars(
                    cmake_file.read_text()
                )
        return cmake_cache_vars

    @classmethod
    def get_cmake_cache_args(cls) -> list[MethodRecord]:
        """Return a list of methodRecord for argparse corresponding to cmake cahe options"""

        cmake_cache_args = []

        def _format_cmake_default(cmake_type, cmake_value):
            """Format CMake variable default value according to cmake type"""
            return (
                False
                if cmake_value == "OFF" and cmake_type == "BOOL"
                else True
                if cmake_value == "ON" and cmake_type == "BOOL"
                else cmake_value
            )

        def _get_argument_action(default_value):
            """Get correct argparse action according to default value"""
            return (
                "store_true"
                if default_value is False
                else "store_false"
                if default_value is True
                else "store"
            )

        cmake_cache_vars = cls.get_cmake_cache_vars()

        for cmake_cache_var in cmake_cache_vars:
            formatted_default = _format_cmake_default(
                cmake_cache_var.get("cmake_type").strip(),
                cmake_cache_var.get("cmake_value").strip(),
            )
            formatted_action = _get_argument_action(formatted_default)

            cmake_cache_args.append(
                MethodRecord(
                    args=[f"--{cmake_cache_var.get('cmake_var')}"],
                    kwargs={
                        "help": _(
                            cmake_cache_var.get("cmake_docstring", "").strip(
                                cmake_cache_var.get("quote")
                            )
                        ),
                        "dest": f"{CMAKE_OPTION_PREFIX}{cmake_cache_var.get('cmake_var')}",
                        "default": formatted_default,
                        "action": formatted_action,
                    },
                )
            )

        return cmake_cache_args

    def build(self):
        """Launch CMake build command with cmake options inferred from mode"""
        self._subprocess_run(
            f'cmake {(self.working_dirs["src"] / "geniac").as_posix()} '
            f"-DCMAKE_INSTALL_PREFIX={self.install_path.as_posix()} "
            f"{' '.join(self.cmake_options)}".split(),
            capture_output=False,
            cmd_name="CMake build",
            check=True,
            cwd=self.working_dirs["build"],
        )

    def install(self):
        """Launch Make installation command in the folder"""


        if self.mode == "singularity" and self.is_sudo:
            def detect_softname_version(softname = None, version_cmd = None):
                softname_cmd = softname + " " + version_cmd
                proc = subprocess.run(softname_cmd.split(), stdout=subprocess.PIPE, encoding="ascii")
                user_version = proc.stdout.strip()
                print(f"Checking that the {softname} version used with sudo privilege is the same as the one defined in the user environment detected during the cmake configuration step.")
                softname_cmd = "sudo -E -S -k " + softname_cmd
                proc = subprocess.run(softname_cmd.split(), stdout=subprocess.PIPE, encoding="ascii")
                root_version = proc.stdout.strip()
                if user_version != root_version:
                    self.error("Installation with sudo privilege. The %s version used by root does not match the one defined in the user environment:\n\tuser version is: %s\n\troot version is: %s\n", softname, user_version, root_version)
                    sys.exit(f"Make sure that '{user_version}' is the first '{softname}' executable to be found in the secure_path option declared in the file /etc/sudoers.")
                else:
                    self.info("Installation with sudo privilege. The '%s' used by root matches the one defined in the user environment. ", root_version)
            detect_softname_version("singularity", "--version")
            detect_softname_version("nextflow", "-version")



        if self.is_sudo:
            self.warning(
                "Detected options require sudo privileges. Will try to launch in sudo mode"
            )
        user_uid=str(os.getuid())
        user_gid=str(os.getgid())
        chown_command='sudo -E -S -k chown -R ' + user_uid + ':' + user_gid + ' ' + (self.working_dirs["build"]).as_posix()
        commands = {
            "make": f"{'sudo -E -S -k ' if self.is_sudo else ''}make",
            "chown": f"{chown_command if self.is_sudo else 'echo OK'}",
            "make install": "make install",
        }
        for command_name, command in commands.items():
            self._subprocess_run(
                command.split(),
                capture_output=(command.split()[0] != "sudo"),
                cmd_name=command_name,
                check=True,
                cwd=self.working_dirs["build"],
            )

    def run(self):
        """Geniac Install entry point"""
        self.build()
        self.install()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""scripts.py: Nextflow scripts parser."""

import re
import typing
from collections import OrderedDict, defaultdict

from geniac.cli.parsers.base import GeniacParser, PathLike

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2021"


class NextflowScript(GeniacParser):
    """Nextflow script file parser"""

    # process flag
    PROCESS_RE = re.compile(r"^ *process +(?P<processName>\w+) *{")
    # process label
    LABEL_RE = re.compile(r"^ *label +['\"](?P<labelName>\w+)['\"] *")
    # process label defined by a variable
    LABEL_VARIABLE_RE = re.compile(r"^ *label +[(](?P<labelParamsValue>.*) \?\: ['\"'](?P<labelName>.*)['\"'][)] *")
    # script flag
    SCRIPT_RE = re.compile(
        r"^ *(?P<startScript>[\"']{3})"
        r"((?P<script>.+)(?<=(?P<endScript>[\"']{3})))? *$"
    )
    # output/input flag
    INOUT_RE=re.compile(r"^[ \t]*(?P<inout>input|output|script):[ \t]*")
    # renvInit invokation
    RENV_INIT_RE = re.compile(r"\s*(?P<renvInit>renv\w+Init)\(['\"'](?P<renvLabel>\w+)['\"']\)")
    RENV_INIT_OUT_RE = re.compile(r"\s*(?P<renvProcess>\w+)\((?P<renvInit>\w+).*\.out\.renvInitDone\)")
    RENV_INIT_INCLUDE_RE = re.compile(r"\s*include\s*{\s*renvInit\s+as\s+(?P<renvInitInclude>\w+)\s*}\s+from\s+['\"'](?P<renvInitFile>.*)['\"']")


    def _read(
        self,
        in_file: typing.Union[typing.IO, typing.BinaryIO],
        in_path: PathLike = None,
        **kwargs,
    ):
        """Load a Nextflow script file into content property

        Args:
            in_file (BinaryIO): path to nextflow script file
            encoding (str): name of the encoding use to decode config files
            in_path (PathLike): path to the input file
            flush_content (bool): flag used to flush previous content before reading
            warnings (bool): flag to turn on/off warning messages
        """
        script_flag = False
        process = ""
        inout = ""
        output_flag = False
        input_flag = False
        # TODO: change process keys to ("processName", filePath)
        self.content["process"] = self.content.get("process") or OrderedDict()
        self.content["renvInitLabel"] = self.content.get("renvInitLabel") or OrderedDict()
        self.content["renvInitOut"] = self.content.get("renvInitOut") or OrderedDict()
        self.content["renvInitInclude"] = self.content.get("renvInitInclude") or OrderedDict()
        for idx, line in enumerate(super()._read(in_file, **kwargs)):
            if match := self.RENV_INIT_RE.match(line):
                values = match.groupdict()
                renvInit = values.get("renvInit")
                renvLabel = values.get("renvLabel")
                self.content["renvInitLabel"][renvLabel] = renvInit
            if match := self.RENV_INIT_OUT_RE.match(line):
                values = match.groupdict()
                renvInit = values.get("renvInit")
                renvProcess = values.get("renvProcess")
                self.content["renvInitOut"][renvInit] = renvProcess
            if match := self.RENV_INIT_INCLUDE_RE.match(line):
                values = match.groupdict()
                renvInitInclude = values.get("renvInitInclude")
                renvInitFile = values.get("renvInitFile")
                self.content["renvInitInclude"][renvInitInclude] = renvInitFile
            if match := self.PROCESS_RE.match(line):
                inout = ""
                input_flag = False
                output_flag = False
                script_flag = False
                values = match.groupdict()
                # If process add it to the process dict
                process = values.get("processName")
                self.content["process"][process] = defaultdict(list)
                # Save the path to the nextflow script for future logs
                self.content["process"][process]["NextflowScriptPath"] = str(in_path)
            if match := self.LABEL_RE.match(line):
                values = match.groupdict()
                label = values.get("labelName")
                self.debug("FOUND label '%s' in process '%s'.", label, process)
                self.content["process"][process]["label"].append(label)
                continue
            if match := self.LABEL_VARIABLE_RE.match(line):
                values = match.groupdict()
                label = values.get("labelName")
                paramsValue = values.get("labelParamsValue")
                self.info("FOUND label '%s' in process '%s' defined using the variable '%s'.", label, process, paramsValue)
                self.content["process"][process]["labelVariable"].append(label)
                self.content["process"][process]["labelVariableParams"].append(paramsValue)
                continue
            # For the moment we append everything into the same list even with conditional nextflow
            # script
            if match := self.SCRIPT_RE.match(line):
                inout = ""
                input_flag = False
                output_flag = False
                values = match.groupdict()
                if process:
                    script_flag = not script_flag
                    if values.get("script"):
                        self.content["process"][process]["script"].append(line.strip())
                continue
            if match := self.INOUT_RE.match(line):
                input_flag = False
                output_flag = False
                script_flag = False
                values = match.groupdict()
                if values.get("inout") == "output":
                    output_flag = True
                if values.get("inout") == "input":
                    input_flag = True
                continue
            # Add to script part if script_flag
            if process and script_flag:
                self.debug("Add line %s to process %s scope for script part.", idx, process)
                self.content["process"][process]["script"].append(line.strip())
                continue
            # Add to output part if output_flag
            if process and output_flag:
                self.debug("Add line %s to process %s scope for output part.", idx, process)
                self.content["process"][process]["output"].append(line.strip())
                continue
            # Add to input part if input_flag
            if process and input_flag:
                self.debug("Add line %s to process %s scope for input part.", idx, process)
                self.content["process"][process]["input"].append(line.strip())
                continue



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Geniac main script"""

import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from gettext import gettext as _
from importlib.resources import files
from logging.config import dictConfig
from sys import argv

from geniac import __version__
from geniac.cli.commands.clean import GeniacClean
from geniac.cli.commands.configs import GeniacConfigs
from geniac.cli.commands.init import GeniacInit
from geniac.cli.commands.install import GeniacInstall
from geniac.cli.commands.lint import GeniacLint
from geniac.cli.commands.options import GeniacOptions
from geniac.cli.commands.recipes import GeniacRecipes
from geniac.cli.commands.test import GeniacTest
from geniac.cli.utils.base import MethodRecord, load_logging_config
import geniac.cli.utils.errorcounter

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

# noinspection PyTypeChecker
_PKG_NAME = "geniac.cli"
_logging_config_path = files(f"{_PKG_NAME}.data.conf").joinpath("logging.json")
_logger = logging.getLogger()
_logging_config = load_logging_config(_logging_config_path)


def setup_logging(log_level=logging.WARNING):
    """Setup basic logging

    Args:
      log_level (int): minimum log_level for emitting messages
    """
    # Setup a default logger
    logging.basicConfig(level=log_level)
    # Update with file handlers defined in _logging_config
    dictConfig(_logging_config)
    logging.captureWarnings(True)


def update_logger(
    logging_config: dict, log_level: int = None, only_stream: bool = False
):
    """Update logging config

    Args:
        logging_config (dict): dict config
        log_level (int): minimum log_level for emitting messages
        only_stream (bool): keep only Stream handlers
    """
    if log_level and logging_config.get("handlers"):
        _logger.setLevel(log_level)
        for handler in _logger.handlers:
            handler.setLevel(log_level)
    if only_stream and logging_config.get("handlers"):
        logging_config["handlers"] = {
            handler: handler_config
            for (handler, handler_config) in logging_config["handlers"].items()
            if handler == "console"
        }

        # And remove them within loggers and root section
        for config_section in ("root", "loggers"):
            if logging_config.get(config_section, {}).get("handlers"):
                logging_config[config_section]["handlers"] = [
                    handler
                    for handler in logging_config[config_section]["handlers"]
                    if handler in logging_config["handlers"]
                ]
    dictConfig(logging_config)


def main():
    """Main entry point"""
    # Initialize logger
    setup_logging()
    GeniacEntryPoint(argv[1:]).main()


class GeniacEntryPoint:
    """Geniac basic entry point"""

    DEFAULT_LOG_LEVELS = {
        "init": logging.INFO,
        "lint": logging.WARNING,
        "install": logging.INFO,
        "options": logging.INFO,
        "clean": logging.INFO,
        "configs": logging.INFO,
        "recipes": logging.INFO,
    }
    DEFAULT_OPTS = (
        MethodRecord(
            args=["--version"],
            kwargs={"action": "version", "version": f"geniac version-{__version__}"},
        ),
        MethodRecord(
            args=["-v", "--verbose"],
            kwargs={
                "dest": "log_level",
                "help": _("Set log level to INFO"),
                "action": "store_const",
                "const": logging.INFO,
            },
        ),
        MethodRecord(
            args=["-vv", "--very-verbose"],
            kwargs={
                "dest": "log_level",
                "help": _("Set log level to DEBUG"),
                "action": "store_const",
                "const": logging.DEBUG,
            },
        ),
        MethodRecord(
            args=["--no-logfiles"],
            kwargs={
                "dest": "only_stream",
                "help": _("Disable generation of log files"),
                "action": "store_true",
            },
        ),
        MethodRecord(
            args=["-c", "--config"],
            kwargs={
                "dest": "config_file",
                "help": _("Path to geniac config file (INI format)"),
                "type": str,
                "metavar": "CONF.INI",
            },
        ),
        MethodRecord(
            args=["-b", "--branch"],
            kwargs={
                "help": _(
                    "Git branch used to clone the project if project path is a correct git URL"
                ),
                "dest": "branch",
                "default": None,
                "metavar": "BRANCH",
            },
        ),
    )
    DEFAULT_ARGS = (
        MethodRecord(
            args=["src_path"],
            kwargs={
                "help": _(
                    "Path or URL to the Nextflow repository on which you want to use geniac. "
                    "The path can be a geniac working directory initialized with "
                    "`geniac init`."
                ),
                "metavar": "SRC_PATH",
                "nargs": "?",
                "default": ".",
            },
        ),
    )
    LINT_ARGS = (
        MethodRecord(
            args=["--conda-check"],
            kwargs={
                "dest": "condaCheck",
                "help": _("Enable check of conda packages with conda"),
                "action": "store_true",
            },
        ),
    )
    INIT_ARGS = (
        MethodRecord(
            args=["-w", "--working-dir"],
            kwargs={
                "help": _(
                    "Path where the geniac working directory will be initialized"
                ),
                "dest": "working_dir",
                "metavar": "WORK_DIR",
            },
        ),
    )
    OPTIONS_ARGS = tuple(
        list(INIT_ARGS)
        + [
            # Put here specific args for options cmd
        ]
    )
    CLEAN_ARGS = tuple(
        list(INIT_ARGS)
        + [
            # Put here specific args for options cmd
        ]
    )
    CONFIGS_ARGS = tuple(
        list(INIT_ARGS)
        + [
            # Put here specific args for options cmd
        ]
    )
    RECIPES_ARGS = tuple(
        list(INIT_ARGS)
        + [
            # Put here specific args for options cmd
        ]
    )
    TEST_ARGS = tuple(
        [
            MethodRecord(
                args=["profile"],
                kwargs={
                    "help": _("Nextflow profile to be tested"),
                    "choices": [
                        "standard",
                        "conda",
                        "multiconda",
                        "singularity",
                        "docker",
                        "path",
                        "multipath",
                    ],
                    "nargs": "+",
                    "metavar": "NXF_PROFILE",
                },
            ),
            MethodRecord(
                args=["--check-cluster"],
                kwargs={
                    "dest": "check_cluster",
                    "help": _("Check also generated cluster executor"),
                    "action": "store_true",
                },
            ),
        ]
        + GeniacInstall.get_cmake_cache_args()
    )
    INSTALL_ARGS = tuple(
        [
            MethodRecord(
                args=["install_path"],
                kwargs={
                    "help": _("PATH where the nextflow repository will be installed"),
                    "metavar": "INSTALL_PATH",
                },
            ),
            MethodRecord(
                args=["--no-post-clean"],
                kwargs={
                    "help": _("Disable cleaning of working folder after installation"),
                    "dest": "no_post_clean",
                    "action": "store_true",
                },
            ),
            MethodRecord(
                args=["--no-pre-clean"],
                kwargs={
                    "help": _("Disable cleaning of working folder before installation"),
                    "dest": "no_pre_clean",
                    "action": "store_true",
                },
            ),
            MethodRecord(
                args=["-m", "--mode"],
                kwargs={
                    "help": _(
                        "Use a predefined mode to build and install the pipeline with all build "
                        "flags turned on [all] or only images [images], docker images [docker], "
                        "singularity images [singularity], Nextflow config files [configs], "
                        "container recipes [recipes]."
                    ),
                    "dest": "mode",
                    "default": None,
                    "choices": [
                        "all",
                        "images",
                        "docker",
                        "podman",
                        "singularity",
                        "singularityfakeroot",
                        "configs",
                        "recipes",
                    ],
                    "metavar": "MODE",
                },
            ),
        ]
        + GeniacInstall.get_cmake_cache_args()
    )

    def __init__(self, args):
        """

        Args:
            args: list of command line arguments passed to the CLI (sys.argv output)
        """
        self.parser = ArgumentParser(
            prog="geniac", description="Geniac Command Line Interface"
        )
        for record in self.DEFAULT_OPTS:
            self.parser.add_argument(*record.args, **record.kwargs)
        self.args = args
        self.parsed_args = None

    def clean_cmd(self):
        """geniac clean subcommand"""
        return GeniacClean(**self.parsed_args, parser=self.parser)

    def configs_cmd(self):
        """geniac configs subcommand"""
        return GeniacConfigs(**self.parsed_args, parser=self.parser)

    def recipes_cmd(self):
        """Geniac recipes subcommand"""
        return GeniacRecipes(**self.parsed_args, parser=self.parser)

    def lint_cmd(self):
        """Geniac Lint subcommand"""
        return GeniacLint(**self.parsed_args, parser=self.parser)

    def install_cmd(self):
        """Geniac install subcommand"""
        return GeniacInstall(**self.parsed_args, parser=self.parser)

    def init_cmd(self):
        """Geniac init subcommand"""
        return GeniacInit(**self.parsed_args, parser=self.parser)

    def options_cmd(self):
        """Geniac options subcommand"""
        return GeniacOptions(**self.parsed_args, parser=self.parser)

    def test_cmd(self):
        """Geniac tests subcommand"""
        return GeniacTest(**self.parsed_args, parser=self.parser)

    def parse_args(self):
        """Parse command line parameters"""

        # Add sub command (lint and conf)
        subparsers = self.parser.add_subparsers(title="commands")

        # Define a parent parser to share common set of arguments
        parent_parser = ArgumentParser(add_help=False)
        for record in self.DEFAULT_OPTS + self.DEFAULT_ARGS:
            parent_parser.add_argument(*record.args, **record.kwargs)

        # Geniac Check
        parser_lint = subparsers.add_parser(
            "lint",
            help=_("Evaluate the compatibility of a Nextflow project with Geniac"),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_lint.set_defaults(func=self.lint_cmd, which="lint")

        # Geniac install
        parser_install = subparsers.add_parser(
            "install",
            help=_("Installation utility of a Nextflow project compatible with geniac"),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_install.set_defaults(func=self.install_cmd, which="install")

        # Geniac options
        parser_options = subparsers.add_parser(
            "options",
            help=_(
                "Call CMake help utility on a Nextflow project compatible with geniac"
            ),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_options.set_defaults(func=self.options_cmd, which="options")

        # Geniac clean
        parser_clean = subparsers.add_parser(
            "clean",
            help=_(
                "Clean build directoty"
            ),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_clean.set_defaults(func=self.clean_cmd, which="clean")

        # Geniac configs
        parser_configs = subparsers.add_parser(
            "configs",
            help=_(
                "Generate config files a Nextflow project compatible with geniac"
            ),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_configs.set_defaults(func=self.configs_cmd, which="configs")

        # Geniac recipes
        parser_recipes = subparsers.add_parser(
            "recipes",
            help=_(
                "Generate container recipes on a Nextflow project compatible with geniac"
            ),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_recipes.set_defaults(func=self.recipes_cmd, which="recipes")

        # Geniac test
        parser_test = subparsers.add_parser(
            "test",
            help=_("Test Nextflow profile(s) generated with geniac"),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_test.set_defaults(func=self.test_cmd, which="test")

        # Geniac init
        parser_init = subparsers.add_parser(
            "init",
            help=_("Initialize a geniac working directory"),
            parents=[parent_parser],
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser_init.set_defaults(func=self.init_cmd, which="init")

        for (argparse_args, parser) in (
            (self.LINT_ARGS, parser_lint),
            (self.INSTALL_ARGS, parser_install),
            (self.OPTIONS_ARGS, parser_options),
            (self.TEST_ARGS, parser_test),
            (self.INIT_ARGS, parser_init),
        ):
            for argparse_arg in argparse_args:
                parser.add_argument(*argparse_arg.args, **argparse_arg.kwargs)

        self.parsed_args = vars(self.parser.parse_args(self.args))

    def main(self):
        """Main entry point allowing external calls"""
        self.parse_args()
        # Initialize logging instance
        update_logger(
            _logging_config,
            self.parsed_args.get("log_level"),
            self.parsed_args.get("only_stream", False),
        )
        if "func" in self.parsed_args:
            _logger.info(
                "Start geniac %s command.", self.parsed_args.get("which", None)
            )
            # Run method is actually the main entry point for any geniac subcommand
            self.parsed_args.get("func")().run()
        else:
            self.parser.print_help()
        _logger.debug("Script ends here.")

        if geniac.cli.utils.errorcounter.error_counter > 0:
            raise SystemExit(1)



if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""check.py: Linter command for geniac"""

import re
import os
import subprocess
from collections import OrderedDict
from inspect import getfullargspec
from pathlib import Path
from shutil import which
import yaml
from yaml.loader import SafeLoader

from geniac.cli.commands.base import GeniacCommand
from geniac.cli.parsers.base import DEFAULT_ENCODING
from geniac.cli.parsers.config import NextflowConfig, NextflowConfigContainer
from geniac.cli.parsers.scripts import NextflowScript

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


class GeniacLint(GeniacCommand):
    """Linter command for geniac"""

    # REGEX check if a string is a valid conda recipe
    CONDA_RECIPES_RE = re.compile(
        r"(?P<recipes>(([\w-]+::[.\w-]+==?[\d.]+.*==?[\w]+) ?)+)"
    )
    # REGEX check if a string from a yml recipe is valid
    CONDA_YML_RECIPES_RE = re.compile(
        r"(?P<recipes>(([.\w-]+==?[\d.]+.*==?[\w]+|[\w-]+::[.\w-]+==?[\d.]+.*==?[\w]+) ?)+)"
    )
    # REGEX check if a string from a yml recipe with pip is valid
    CONDA_YML_PIP_RECIPES_RE = re.compile(
        r"(?P<recipes>([.\w-]+==?[\d.]))"
    )
    # REGEX to check if a string is a path for yml or yaml file
    CONDA_PATH_RE = re.compile(
        #r"(?P<nxfvar>\${(baseDir|projectDir)})/(?P<basepath>[/\w]+\.(?P<ext>yml|yaml))"
        r"(?P<nxfvar>\${(baseDir|projectDir)})/(?P<basepath>.+\.(?P<ext>yml|yaml))"
    )
    # REGEX to check if a string is a path for renv.lock file
    RENV_LOCKFILE_PATH_RE = re.compile(
        r"(?P<nxfvar>\${(baseDir|projectDir)})/(?P<basepath>[/\w]+\.(?P<ext>lock))"
    )
    # REGEX to check if install cmake directive has been correctly added in the main CMakeLists.txt
    INSTALL_MAIN_CMAKE_RE = re.compile(
        r"install\([\s\w_${}\-/=]*DESTINATION +"
        r"(?P<destination>\${CMAKE_INSTALL_PREFIX}/\${pipeline_dir}/bin/fromSource)[\s)]"
    )
    # REGEX to check if install cmake directive has been correctly added in the
    # module CMakeLists.txt
    INSTALL_MODULE_CMAKE_RE = re.compile(
        r"install\s*\([\s\w_${}\-/=]*DESTINATION +[\s\w_${}\-/=]*\)"
    )
    # REGEX to check if ExternalProject cmake directive has been correctly added in the main
    # CMakeLists.txt
    PROJECT_ADD_MAIN_CMAKE_RE_TEMP = (
        r"ExternalProject_Add\(\s.*\s*SOURCE_"
        r"DIR +(\$\{{pipeline_source_dir\}}/modules/fromSource|"
        r"\$\{{CMAKE_CURRENT_SOURCE_DIR\}})/{label}\n"
    )
    # REGEX to check if a dependency has been correctly added in a singularity
    # recipe
    SINGULARITY_DEP_RE_TEMP = (
        r"\%files[\/\w.\s]*\s+(?P<mydep>{tool}/{dependency} +[\/\w.]+{dependency})"
    )
    # REGEX to check if a dependency has been correctly added in a docker
    # recipe
    DOCKER_DEP_RE_TEMP = r"(ADD|COPY) +{tool}/{dependency} [\/\w.]+{dependency}"
    

    # Name of config sections used in this class
    TREE_SUFFIX = "tree"
    PROJECT_WORKFLOW = "project.workflow"
    GENIAC_PARAMS = "geniac.lint"
    GENIAC_DIRS = "geniac.lint.directories"
    GENIAC_CONFIG_FILES = "geniac.lint.generated.config"
    GENIAC_CHECK_CONFIG = "geniac.lint.expected.config"

    def __init__(self, src_path, *args, **kwargs):
        """Init flags specific to GCheck command"""
        super().__init__(*args, src_path=src_path, **kwargs)
        self._project_tree = self._format_tree_config()
        self._labels_from_folders = OrderedDict()
        self._labels_from_configs = OrderedDict()
        self._processes_from_workflow = OrderedDict()
        self._renvinitlabel_from_workflow = OrderedDict()
        self._renvinitout_from_workflow = OrderedDict()
        self._renvinitinclude_from_workflow = OrderedDict()
        self._labels_from_workflow = []
        self._labels_all = []
        self._nxf_config_container = NextflowConfigContainer()

    @property
    def project_tree(self):
        """Formatted tree configuration"""
        return self._project_tree

    @property
    def nxf_config_container(self):
        """Namespace for Nextflow configuration files"""
        return self._nxf_config_container

    @property
    def labels_from_folders(self):
        """Geniac labels from Nextflow folders"""
        return self._labels_from_folders

    @labels_from_folders.setter
    def labels_from_folders(self, value: dict):
        """Merge geniac labels from Nextflow folders"""
        self._labels_from_folders |= value

    @property
    def labels_from_configs(self):
        """Geniac labels from Nextflow configs"""
        return self._labels_from_configs

    @labels_from_configs.setter
    def labels_from_configs(self, value: dict):
        """Merge geniac labels from Nextflow configs"""
        self._labels_from_configs |= value

    @property
    def labels_from_geniac_config(self):
        """Geniac labels from Nextflow folders"""
        return list(dict.fromkeys(list(self.labels_from_configs.get("geniac", []))))

    @property
    def labels_from_process_config(self):
        """Process config labels from Nextflow folders"""
        return list(dict.fromkeys(list(self.labels_from_configs.get("process", []))))

    @property
    def processes_from_workflow(self):
        """Workflow labels from Nextflow folders"""
        return self._processes_from_workflow

    @processes_from_workflow.setter
    def processes_from_workflow(self, value: dict):
        """Merge geniac labels from Nextflow configs"""
        self._processes_from_workflow |= value

    @property
    def labels_from_workflow(self):
        """Workflow labels from Nextflow folders"""
        # Init labels list if empty
    
        labels = list(
            dict.fromkeys(
                [
                    label
                    for process, process_scope in self.processes_from_workflow.items()
                    for label in process_scope["label"]
                    if label is not None
                ]
            )
        )
        self._labels_from_workflow += set(labels + self._labels_from_workflow)
        return self._labels_from_workflow

    @property
    def labels_all(self):
        """Gather labels from Nextflow folders and geniac tools"""
        # Init labels all if empty
        if not self._labels_all:
            self._labels_all = list(
                dict.fromkeys(
                    [
                        label
                        for folder, labels in self.labels_from_folders.items()
                        for label in labels
                    ]
                    + self.labels_from_geniac_config
                    + ["onlyLinux"]
                )
            )
        return self._labels_all

    def _get_current_files(self, config_tree: dict, tree_section: str):
        """
        Get current file list from a specific section

        Args:
            config_tree:
            tree_section:

        Returns:

        """
        dir_path = config_tree.get(tree_section).get("path")
        recursive_flag = config_tree.get(tree_section).get("recursive")
        excluded_files = config_tree.get(tree_section).get("excluded_files")
        self.debug(
            "Browse current files in %s directory%s",
            dir_path,
            " recursively" if recursive_flag else "",
        )
        return (
            sorted(
                [
                    _
                    for _ in dir_path.glob("**/*" if recursive_flag else "*")
                    if _ not in excluded_files and not _.is_dir()
                ]
            )
            if dir_path.exists()
            else ()
        )

    def _format_tree_config(self):
        """Format configuration tree from ini config

        Returns:
            config_tree (dict)
        """
        config_tree = OrderedDict(
            (
                tree_section.removeprefix(self.TREE_SUFFIX + "."),
                {
                    # Is the folder required ?
                    "required": self.default_config.getboolean(tree_section, "required")
                    if self.default_config.has_option(tree_section, "required")
                    else False,
                    # Is the folder recommended ?
                    "recommended": self.default_config.getboolean(
                        tree_section, "recommended"
                    )
                    if self.default_config.has_option(tree_section, "recommended")
                    else False,
                    # Should we analyze files and sub directories recursively ?
                    "recursive": self.default_config.getboolean(
                        tree_section, "recursive"
                    )
                    if self.default_config.has_option(tree_section, "recursive")
                    else False,
                    # Path to the folder
                    "path": Path(self.default_config.get(tree_section, "path"))
                    if self.default_config.get(tree_section, "path")
                    else Path(self.src_path),
                    # Path(s) to mandatory file(s)
                    "required_files":  sorted(
                        list(
                            set(
                                [
                                    path
                                    for path in self.get_config_path(tree_section,
                                                                     "mandatory",
                                                                     lazy_glob=True
                                                                     )
                                    if path not in self.get_config_path(tree_section,
                                                                        "excluded",
                                                                        lazy_glob=True
                                                                        )
                                ]
                            )
                        )
                    ),
                    # Path(s) to optional file(s)
                    "optional_files": sorted(
                        list(
                            set(
                                [
                                    path
                                    for path in self.get_config_path(
                                        tree_section, "optional"
                                    )
                                    if path
                                    not in self.get_config_path(
                                        tree_section, "mandatory", lazy_glob=True
                                    )
                                ]
                            )
                        )
                    ),
                    # Path(s) to file(s) excluded from the analysis
                    "excluded_files": self.get_config_path(tree_section, "excluded"),
                    # Path(s) to file(s) excluded from the analysis
                    "prohibited_files": self.get_config_path(
                        tree_section, "prohibited"
                    ),
                },
            )
            for tree_section in self.get_config_subsection(self.TREE_SUFFIX)
        )
        return OrderedDict(
            (
                tree_section,
                {
                    # Get a list all the files in the folder
                    "current_files": (
                        self._get_current_files(config_tree, tree_section)
                    ),
                    **section,
                },
            )
            for tree_section, section in config_tree.items()
        )

    def check_tree_folder(self):
        """Check the directory in order to set the flags"""
        self.info("Checking tree structure of %s.", self.src_path)
        self.debug(
            "Sections parsed from config file: %s.", self.default_config.sections()
        )

        for tree_section, section in self.project_tree.items():
            for msg in ("\n", f"Folder {tree_section}"):
                self.debug(msg)

            # Is the actual folder required
            required = section.get("required")
            # Is the actual folder recommended
            recommended = section.get("recommended")
            # Path to the sub directory analyzed
            path = section.get("path")
            # List of required files requested in configuration file(s)
            required_files = section.get("required_files")
            # List of optional files requested in configuration file(s)
            optional_files = section.get("optional_files")
            # List of files actually present in the directory
            current_files = section.get("current_files")

            for msg in (
                f"required: {required}",
                f"path: {path}",
                f"expected files: {required_files}",
                f"optional files: {optional_files}",
                f"excluded files: {section.get('excluded_files')}",
                f"current files: {current_files}",
            ):
                self.debug(msg)

            # If folder exists and is not empty (excluded files are ignored)
            if path:
                is_src_path = path.resolve() == self.src_path.resolve()
                formatted_path = (
                    path.relative_to(self.src_path)
                    if not is_src_path
                    else self.src_path.relative_to(Path.cwd())
                    if self.src_path.is_relative_to(Path.cwd())
                    else self.src_path
                )
                if required and not path.exists():
                    extra_msg = (
                        " Add it to your project if you want your "
                        "workflow to be compatible with geniac tools."
                        if not is_src_path
                        else ""
                    )
                    self.critical(
                        "Directory %s does not exist.%s", formatted_path, extra_msg
                    )
                elif recommended and not path.exists():
                    self.warning(
                        "Directory %s does not exist. It is recommended to have one in your "
                        "project.",
                        formatted_path,
                    )

            # Trigger an error if a mandatory file is missing
            for file in required_files:
                # If the folder is actually required but the required file is not
                # present or if the folder is recommended and non empty
                if (required or (recommended and current_files)) and (
                    file not in current_files
                ):
                    self.error(
                        "File %s is missing. Add it to your project if you want to be compatible "
                        "with geniac.",
                        file.relative_to(self.src_path),
                    )

            # Trigger a warning if an optional file is missing
            for file in optional_files:
                # If the folder is actually required but the optional file is not
                # present
                if required and file not in current_files:
                    self.warning(
                        "Optional file %s does not exist. It is recommended to have one in your "
                        "project.",
                        file.relative_to(self.src_path),
                    )

    def get_processes_from_workflow(self):
        """Parse workflow file(s)

        Returns:
            labels_from_main (dict): dictionary of processes in the main nextflow file
        """
        script = NextflowScript(src_path=self.src_path)

        geniac_dir = self.src_path / "geniac"

        # Link config path to their method
        script_paths = OrderedDict(
            (
                f"{config_key}_{index}",
                path,
            )
            for config_key in self.default_config.options(GeniacLint.PROJECT_WORKFLOW)
            for index, path in enumerate(
                self.get_config_path(GeniacLint.PROJECT_WORKFLOW, config_key)
            )
            if not path.is_relative_to(geniac_dir)
        )

        for _, script_path in script_paths.items():
            if script_path.exists():
                script.read(script_path)
            else:
                self.error(
                    "Workflow script %s does not exist.",
                    script_path.relative_to(self.src_path),
                )

        # Check if there is processes without label in the actual workflow
        # fmt: off
        for process in (processes := script.content.get("process")):
            if not processes.get(process).get("label"):
                process_path = Path(
                    processes.get(process).get('NextflowScriptPath')
                ).relative_to(self.src_path)
                self.error(
                    "Process %s in %s does not have any label.",
                    process, process_path
                )
        # fmt: on

        self.processes_from_workflow = script.content.get("process", OrderedDict())
        self.renvinitlabel_from_workflow = script.content.get("renvInitLabel", OrderedDict())
        self.renvinitout_from_workflow = script.content.get("renvInitOut", OrderedDict())
        self.renvinitinclude_from_workflow = script.content.get("renvInitInclude", OrderedDict())

    def _check_geniac_config(
        self, config: NextflowConfig, conda_check: bool = False
    ) -> list:
        """Check the content of params scope in a geniac config file

        Args:
            conda_check (bool): trigger conda search calls with conda recipes in params.geniac.tools

        Returns:
            labels_geniac_tools (list): list of geniac tool labels in params.geniac.tools
        """
        # list of label declared in params.geniac.tools
        labels_geniac_tools = []
        # list of tools used as variable in other params.geniac.tools
        labels_variables_tools = []

        # Check parameters according to their default values
        config.check_config_scope("params")

        # REGEX to check if a label uses information from another label
        # recipe
        LABEL_USED_BY_A_LABEL = r"\s*\$\{params\.geniac\.tools\.(?P<usedLabel>.+)\}\s*"

        # Check if conda command exists
        if conda_check:
           cmd = which("conda")
           if cmd == None:
              self.error(
                  "Conda is not available in your path. Geniac will not check if tool "
                  "recipes are correct. Add conda in your PATH: export PATH=/path/to/conda/bin:$PATH"
              )
           else:
              self.info("conda is in the PATH")

        # Check each label in params.geniac.tools
        self.info(
            "Checking conda recipes in params.geniac.tools."
            if conda_check
            else "Checking of conda recipes turned off."
        )
        geniac_tools_list = config.get("params.geniac.tools", OrderedDict()).items()
        conda_env_name = []
        for label, value in geniac_tools_list:
            # If the len(value) equals 1, then this is a standard label,
            # otherwise, it contains scopes such as the label for renv.
            if len(value) == 1:
                [recipe] = value
                (recipe, n_sub) = re.subn(LABEL_USED_BY_A_LABEL , "", recipe)
                if n_sub > 0:
                    [from_labels] = value
                    from_labels = from_labels.split()
                    from_labels = list(filter(re.compile(LABEL_USED_BY_A_LABEL).match, from_labels))
                    for new_used_label in from_labels:
                        new_used_label = re.match(LABEL_USED_BY_A_LABEL, new_used_label)
                        new_used_label = new_used_label.group('usedLabel')
                        labels_variables_tools.append(new_used_label)
                    self.info("The label '%s' defined in the geniac.config file uses information from the labels %s.", label, labels_variables_tools)
                labels_geniac_tools.append(label)
                if len(recipe) != 0:
                    # If the tool value is a path to an environment file (yml or yaml ext),
                    # check if the path exists
                    if match := GeniacLint.CONDA_PATH_RE.search(recipe):
                        if conda_path := Path(self.src_path / match.groupdict().get("basepath")):
                            if conda_path.exists():
                                with open(conda_path, 'r') as yml_f:
                                    try:
                                        yml_content = list(yaml.load_all(yml_f, Loader=SafeLoader))[0]
                                    except yaml.YAMLError as exception:
                                        self.error("The file '%s' is not correctly formatted. Check that the YAML syntax is correct.", conda_path.relative_to(self.src_path))
                                        raise exception
                                    if not 'name' in yml_content:
                                        self.error(
                                            "Conda file %s related to %s tool does not have a name entry for the conda environment. For example, add 'name: someValue_env' in the file %s.",
                                            conda_path.relative_to(self.src_path),
                                            label,
                                            conda_path.relative_to(self.src_path)
                                        )
                                    else:
                                        conda_env_name.append(yml_content['name'])
                                    if 'dependencies' in yml_content:
                                        for dep_in_yml in yml_content['dependencies']:
                                            if type(dep_in_yml) is str and dep_in_yml != 'pip':
                                                match = GeniacLint.CONDA_YML_RECIPES_RE.match(dep_in_yml)
                                                if not match:
                                                   self.error(
                                                       "In the file '%s', the value '%s' of '%s' tool does not follow the pattern "
                                                       '"condaChannelName::softName=version=buildString" or "softName=version=buildString".',
                                                       conda_path.relative_to(self.src_path),
                                                       dep_in_yml,
                                                       label
                                                   )
                                            else:
                                                if type(dep_in_yml) is dict and 'pip' in dep_in_yml:
                                                    for pip_tool in dep_in_yml['pip']:
                                                        match = GeniacLint.CONDA_YML_PIP_RECIPES_RE.match(pip_tool)
                                                        if not match:
                                                           self.error(
                                                               "In section 'pip' of the file '%s', the value '%s' of '%s' tool does not follow the pattern "
                                                               '"softName==version".',
                                                               conda_path.relative_to(self.src_path),
                                                               pip_tool,
                                                               label
                                                           )
                                # Test that the name of the conda recipe (yml or yaml file) is consistent with the label
                                if label != os.path.splitext(os.path.basename(conda_path))[0]:
                                        self.error(
                                                "In the file 'conf/geniac.config', the label '%s' uses the custom conda recipe '%s'. Rename the recipe into '%s' to match your label.",
                                                label,
                                                recipe,
                                                os.path.dirname(recipe) + "/" + label + ".yml")
                            else:
                                self.error(
                                    "Conda file %s related to %s tool does not exist.",
                                    conda_path.relative_to(self.src_path),
                                    label,
                                )
                    # Elif the tool value is a conda recipe
                    #elif match := GeniacLint.CONDA_RECIPES_RE.match(recipe):
                    else:
                        for conda_recipe in recipe.split(" "):
                            match = GeniacLint.CONDA_RECIPES_RE.match(conda_recipe)
                            if not match:
                               self.error(
                                   "Value %s of %s tool does not follow the pattern "
                                   '"condaChannelName::softName=version=buildString".',
                                   conda_recipe,
                                   label,
                               )
                        if conda_check:
                            # Check if the recipes exists in the actual OS with conda search
                            try:
                                conda_search = subprocess.run(
                                    ["conda", "search", conda_recipe],
                                    capture_output=True,
                                    check=True,
                                )
                            except subprocess.CalledProcessError:
                                self.error(
                                    "Conda search command returned non-zero exit status for the recipe "
                                    "%s[%s]. Either conda is not available or the recipe does not link "
                                    "to an existing package or build. Check if the requested build is "
                                    "still available on conda with the following command:"
                                    "\n\t> conda search %s.",
                                    conda_recipe,
                                    label,
                                    conda_recipe,
                                )
                            else:
                                self.debug("Conda search output:\n%s", conda_search.stdout)
            else:
                if bool(re.match(r"^renv.*", label)):
                    # renv.lock file must be located in a folder with the name of the label
                    # we will test if the file exists later 
                    renvLockfile = "${projectDir}/recipes/dependencies/" + label + "/renv.lock"
                    #if list(self.processes_from_workflow.keys()).count(label + 'Init') < 1:
                    # A dedicated renv<label>Init process takinf the <lavel> as argument must be called
                    # The renv<label>Init process is used to create conda env when needed
                    if list(self.renvinitlabel_from_workflow.keys()).count(label) < 1:
                        self.error("The workflow has the process with label '%s' with R packages using renv. The process '%s('%s')' is missing. It must be called in your workflow. See https://geniac.readthedocs.io/en/latest/renv.html.", label, label + 'Init', label)
                    else:
                        # Check that renv<label>Init process is correctly named
                        if self.renvinitlabel_from_workflow[label] != label + 'Init':
                            self.error("The workflow has the process with label '%s' with R packages using renv. The process %s has been called in the workflow. Use the syntax %s instead (i.e. concatenate the label name \'%s\' with 'Init' suffix). See https://geniac.readthedocs.io/en/latest/renv.html.", label, '\'' +  self.renvinitlabel_from_workflow[label] + '(\'' + label + '\')' +'\'', '\'' + label + 'Init(' + '\'' + label + '\')\'', label)
                        else:
                            # Check that the module renvInit has been included
                            if list(self.renvinitinclude_from_workflow.keys()).count(self.renvinitlabel_from_workflow[label])  < 1:
                                self.error("The workflow has the process with label '%s' with R packages using renv. The include directive with the renvInit module is missing. Add a line such as: include { renvInit as %s } from \'./nf-modules/local/process/renvInit\'. See https://geniac.readthedocs.io/en/latest/renv.html.", label, label + 'Init')
                            else:
                                renvinit_file = self.src_path / str(self.renvinitinclude_from_workflow[self.renvinitlabel_from_workflow[label]] + ".nf")
                                # Test that the renvInit module exist
                                if not renvinit_file.exists():
                                    self.error("The workflow has the process with label '%s' with R packages using renv. The module %s has been included but it does not exist. See https://geniac.readthedocs.io/en/latest/renv.html.", label, self.renvinitinclude_from_workflow[self.renvinitlabel_from_workflow[label]] + ".nf")
                                    
                            # Check that the output from renv<label>Init as been used by the process which possibl needs the conda env
                            if list(self.renvinitout_from_workflow.keys()).count(self.renvinitlabel_from_workflow[label]) < 1:
                                self.error("The workflow has the process with label '%s' with R packages using renv. The process with the label \'%s\' must be called with the argument %s. See https://geniac.readthedocs.io/en/latest/renv.html.", label, label, '\'' +  self.renvinitlabel_from_workflow[label] + '.out.renvInitDone\'')
                            else:
                                # Check that the process exist
                                process_with_init = self.renvinitout_from_workflow[self.renvinitlabel_from_workflow[label]]
                                if list(self.processes_from_workflow.keys()).count(process_with_init) < 1:
                                    self.error("The workflow has the process with label '%s' with R packages using renv. The argument %s has been passed to the process \'%s\' which has not been defined. See https://geniac.readthedocs.io/en/latest/renv.html.", label, '\'' +  self.renvinitlabel_from_workflow[label] + '.out.renvInitDone\'', process_with_init)
                                else:
                                    # Check that correct process is used
                                    if self.processes_from_workflow[process_with_init].get('label').count(label) < 1:
                                        self.error("The workflow has the process with label '%s' with R packages using renv. The argument %s has been passed to the process \'%s\' but this process does not have any label \'%s\'. See https://geniac.readthedocs.io/en/latest/renv.html.", label, '\'' +  self.renvinitlabel_from_workflow[label] + '.out.renvInitDone\'', process_with_init, label)

                    for scope in ['yml', 'env', 'bioc']:
                        if value.get(scope) == None:
                            self.error("In the renv label '%s', the scope '%s' is missing.", label, scope)
                        else:
                            if scope == 'yml':
                                [renvYml] = value.get('yml')

                                if match := GeniacLint.CONDA_PATH_RE.search(renvYml):
                                    if (
                                        conda_path := Path(
                                            self.src_path / match.groupdict().get("basepath")
                                        )
                                    ) and not conda_path.exists():
                                        self.error(
                                            "Conda file %s related to the renv %s tool does not exist.",
                                            conda_path.relative_to(self.src_path),
                                            label,
                                        )

                    if match := GeniacLint.RENV_LOCKFILE_PATH_RE.search(renvLockfile):
                        if (
                            dep_path := Path(
                                self.src_path / match.groupdict().get("basepath")
                            )
                        ) and not dep_path.exists():
                            self.error("There is no 'recipes/dependencies/%s/renv.lock' file for the renv '%s' tool. You must add the renv.lock file.", label, label)

        # Check that all yml file for conda recipe have a different "name" value
        duplicated_conda_env_name = find_duplicates(conda_env_name)
        if duplicated_conda_env_name:
            self.error("You have different conda recipes in the yml files located in the folder 'recipes/conda' which use the name conda environment name: %s. Use a different conda environment name for each recipe.", duplicated_conda_env_name)

        labels_not_present = set(labels_variables_tools) - set(labels_geniac_tools)
        if labels_not_present:
            self.error("The tools %s used as variables in other tools are not found in the conf/geniac.config.", sorted(list(labels_not_present)))

        return labels_geniac_tools

    def _check_process_config(self, config: NextflowConfig, config_path):
        """Check the content of a process config file

        Returns:
            labels_process (list): list of process labels in params.process with withName
        """
        # Check parameters according to their default values
        config.check_config_scope("process")

        # For each process used with withName selector, check their existence in the
        # workflow
        for config_process in config.get("process", OrderedDict()).get(
            "withName", OrderedDict()
        ):
            if config_process not in self.processes_from_workflow:
                self.error(
                    "withName:%s is defined in %s file but the process %s is not used anywhere.",
                    config_process,
                    config_path.relative_to(self.src_path),
                    config_process,
                )

        # Return list of labels defined with withLabel selector in the process.config file
        return list(config.get("process", OrderedDict()).get("withLabel", []))

    def _check_nextflow_config(
        self,
        config: NextflowConfig,
        config_path,
        default_config_paths: list = (),
        default_geniac_files_paths: list = (),
    ):
        """Check the content of a nextflow config file

        Args:
            config_path:
            default_geniac_files_paths (list):
            default_config_paths (list):
        """
        include_config_paths = [
            self.src_path / Path(include_path)
            for include_path in config.get("includeConfig", [])
        ]
        profile_config_paths = [
            self.src_path / Path(conf_path)
            for conf_profile in config.get("profiles", OrderedDict())
            for conf_path in config.get("profiles", OrderedDict())
            .get(conf_profile, {})
            .get("includeConfig", {})
        ]
        for default_config_path in default_config_paths + default_geniac_files_paths:
            # We do not check if the path corresponds to nextflow.config path
            # Check if config files are included
            if default_config_path != config_path and (
                default_config_path not in include_config_paths + profile_config_paths
            ):
                msg = (
                    f"Main configuration file "
                    f"{config_path.relative_to(self.src_path)} does not "
                    f"include configuration file "
                    f"{default_config_path.relative_to(self.src_path)}."
                )
                # Trigger a warning if optional file. Otherwise trigger an error
                if default_config_path in self.get_config_path(
                    ".".join([GeniacLint.TREE_SUFFIX, "conf"]), "optional"
                ):
                    self.warning(msg)
                else:
                    self.error(msg)

    def _check_base_config(self, config: NextflowConfig):
        """Check the content of a base config file"""
        self.debug("Check content of base config file without analyzing geniac scope")
        config.check_config_scope("params", skip_nested_scopes=["geniac"])

    def get_labels_from_config_files(self):
        """Check the structure of the repo

        Returns:
            labels_geniac_tools (list): list of geniac tool labels in params.geniac.tools
            labels_process (list): list of process labels in params.process with withName
        """
        # Link config path to their method
        project_config_scopes = OrderedDict(
            (
                default_config_name,
                {
                    "expected": True,
                    "path": self.get_config_path(
                        GeniacLint.GENIAC_CHECK_CONFIG,
                        default_config_name,
                        single_path=default_config_name,
                    ),
                    "check_config": getattr(
                        self, f"_check_{default_config_name}_config", None
                    ),
                },
            )
            for default_config_name in self.get_config_section_items(
                GeniacLint.GENIAC_CHECK_CONFIG
            )
            if default_config_name != "all"
        )
        project_config_scopes |= OrderedDict(
            (
                default_config_path.stem,
                {
                    "expected": default_config_path.stem
                    in self.get_config_section_items(GeniacLint.GENIAC_CHECK_CONFIG),
                    "path": default_config_path,
                    "check_config": getattr(
                        self, f"_check_{default_config_path.stem}_config", None
                    ),
                },
            )
            for default_config_path in self.get_config_path(
                GeniacLint.GENIAC_CHECK_CONFIG, "all", single_path=False
            )
            if default_config_path.stem not in project_config_scopes
        )

        # Generate path to configuration files produced by geniac
        generated_geniac_config_paths = [
            self.get_config_path(
                GeniacLint.GENIAC_CONFIG_FILES, geniac_config_file, single_path=True
            )
            for geniac_config_file in self.default_config.options(
                GeniacLint.GENIAC_CONFIG_FILES
            )
        ]

        expected_geniac_config_paths = [
            config_scope.get("path")
            for config, config_scope in project_config_scopes.items()
            if config_scope.get("expected")
        ]

        # For each Nextflow configuration file analyzed
        for config_key, project_config_scope in project_config_scopes.items():
            project_config_path = project_config_scope["path"]
            config_method = project_config_scope["check_config"]
            config_args = {
                "config_path": project_config_path,
                "default_config_paths": expected_geniac_config_paths,
                "default_geniac_files_paths": generated_geniac_config_paths,
                "conda_check": self.default_config.getboolean(
                    self.GENIAC_PARAMS, "condaCheck"
                ),
            }

            # If the project config file does not exists and does not belong to default
            # geniac files
            if not project_config_path.exists():
                if project_config_path not in expected_geniac_config_paths:
                    self.error(
                        "Nextflow config file %s does not exist.",
                        project_config_path.relative_to(self.src_path),
                    )
                continue

            # Read the Nextflow configuration file
            nxf_config = NextflowConfig(src_path=self.src_path, config_file=self.config_file)
            nxf_config.read(
                project_config_path,
                warnings=config_key
                not in self.default_config.options(GeniacLint.GENIAC_CHECK_CONFIG),
            )
            self.nxf_config_container.append(nxf_config)

            if config_method:
                self.info(
                    "Checking Nextflow configuration file. %s",
                    project_config_path.relative_to(self.src_path),
                )
                self.labels_from_configs[config_key] = config_method(
                    **{
                        arg: config_args.get(arg)
                        for arg in getfullargspec(config_method).args
                        if arg != "self"
                    }
                    | {"config": nxf_config}
                )

    def _get_labels_from_modules_dir(self, modules_tree: dict):
        """Get geniac labels from modules directory"""
        labels_from_modules = []
        modules_dir = modules_tree.get("path")
        main_cmake_lists = modules_dir / "CMakeLists.txt"
        if not main_cmake_lists.exists():
            # Output an error if modules directory is not empty
            if any(modules_dir.iterdir()):
                self.error(
                    "Folder %s requires a CMakeLists.txt file in order to automatically "
                    "build containers.",
                    modules_dir.relative_to(self.src_path),
                )
            return []

        with open(main_cmake_lists, encoding=DEFAULT_ENCODING) as cmake_file:
            main_cmake_lists_content = cmake_file.read()

        for module_dir in [
            module for module in modules_dir.iterdir() if module.is_dir()
        ]:
            # If child correspond to a folder and the name of this folder is linked to
            # an existing bash script
            # If the actual file is not the main cmakelists file, it should correspond to a module
            module_name = module_dir.stem
            cmakelists_child = module_dir / "CMakeLists.txt"
            labels_from_modules += [module_name]

            if cmakelists_child.exists():
                self.debug("Found module directory with label %s.", module_name)
                # Parse the CMakeLists.txt file to see if the label is correctly defined
                check_main_cmlist_reg = re.compile(
                    GeniacLint.PROJECT_ADD_MAIN_CMAKE_RE_TEMP.format(label=module_name)
                )

                with open(
                    cmakelists_child, encoding=DEFAULT_ENCODING
                ) as cmake_module_file:
                    module_cmake_lists_content = cmake_module_file.read()

                # First look if the is correctly added within the main CMakeLists.txt file
                if check_main_cmlist_reg.search(main_cmake_lists_content):
                    self.debug(
                        "Module %s correctly added within %s.",
                        module_name,
                        main_cmake_lists.relative_to(self.src_path),
                    )
                else:
                    self.error(
                            "Module %s not added with ExternalProject_Add directive within %s file, or the ExternalProject_Add is not correctly formatted. It should look like this:\nExternalProject_Add(\n\t%s\n\tSOURCE_DIR ${CMAKE_CURRENT_SOURCE_DIR}/%s\n\tCMAKE_ARGS\n\t-DCMAKE_INSTALL_PREFIX=${CMAKE_BINARY_DIR}/externalProject/bin)",
                        module_name,
                        main_cmake_lists.relative_to(self.src_path),
                        module_name,
                        module_name,
                    )

                # Then look in the CMakeLists.txt if install DESTINATION is correct
                if GeniacLint.INSTALL_MAIN_CMAKE_RE.search(main_cmake_lists_content):
                    self.debug(
                        "Module %s correctly setup in %s to install tools inside "
                        "${projectDir}/bin/fromSource",
                        module_name,
                        main_cmake_lists.relative_to(self.src_path),
                    )
                else:
                    self.error(
                        "DESTINATION in '%s' is not set to '${projectDir}/bin/fromSource'. Please "
                        "update DESTINATION section in this file to "
                        '"DESTINATION ${CMAKE_INSTALL_PREFIX}/${pipeline_dir}/bin/fromSource."',
                        main_cmake_lists.relative_to(self.src_path),
                    )

                # Then look in the module cmakeLists.txt if install directive has been set
                if GeniacLint.INSTALL_MODULE_CMAKE_RE.search(
                    module_cmake_lists_content
                ):
                    self.debug("Module %s have an install directive", module_name)
                else:
                    self.warning(
                        "Module %s doesn't setup an install directive within %s file. "
                        "It needs one in order to install sources inside this module.",
                        module_name,
                        cmakelists_child.relative_to(self.src_path),
                    )

        return OrderedDict([("modules", labels_from_modules)])

    @staticmethod
    def _get_labels_from_conda_dir(conda_tree):
        """Get geniac labels from conda recipes"""
        labels_from_recipes = []

        for recipe_child in conda_tree.get("current_files", []):
            labels_from_recipes += [recipe_child.stem]

        return OrderedDict([("conda", labels_from_recipes)])

    @staticmethod
    def _get_labels_from_singularity_dir(singularity_tree):
        """Get geniac labels from singularity recipes"""
        labels_from_recipes = []

        for recipe_child in singularity_tree.get("current_files", []):
            labels_from_recipes += [recipe_child.stem]

        return OrderedDict([("singularity", labels_from_recipes)])

    @staticmethod
    def _get_labels_from_docker_dir(docker_tree):
        """Get geniac labels from docker recipes"""
        labels_from_recipes = []

        for recipe_child in docker_tree.get("current_files", []):
            labels_from_recipes += [recipe_child.stem]

        return OrderedDict([("docker", labels_from_recipes)])

    def _check_dependencies_dir(
        self,
        dependencies_tree: dict,
        docker_tree: dict = None,
        singularity_tree: dict = None,
    ):
        """

        Args:
            dependencies_tree:
            singularity_tree:
            docker_tree:

        Returns:

        """
        dependencies_dir = dependencies_tree.get("path")

        for dependency_path in dependencies_tree.get("current_files", []):
            # The dependency should be inside a sub folder (recipes/dependencies/tool_name/dep.ext)
            tool_name = dependency_path.parent.name
            if (
                tool_name == "dependencies"
                or not dependency_path.parent.parent.resolve().samefile(
                    dependencies_dir.resolve()
                )
            ):
                self.error(
                    "Dependency %s can't be used for container recipes. It should be located "
                    "inside a custom folder with the name corresponding to the container recipe "
                    "file.",
                    dependency_path.relative_to(self.src_path),
                )
                continue

            for recipe_type, recipe_ext, tree in (
                ("singularity", ".def", singularity_tree),
                ("docker", ".Dockerfile", docker_tree),
            ):
                recipe_files = tree.get("current_files", [])
                recipe_flag = False

                # Check if the file is used in recipe files
                for recipe_path in recipe_files:
                    if recipe_path.suffix == recipe_ext:
                        dependency_reg = re.compile(
                            getattr(
                                GeniacLint, f"{recipe_type.upper()}_DEP_RE_TEMP", ""
                            ).format(dependency=dependency_path.name, tool=tool_name)
                        )
                        with open(
                            recipe_path, encoding=DEFAULT_ENCODING
                        ) as recipe_file:
                            recipe_flag = (
                                True
                                if dependency_reg.search(recipe_file.read())
                                else recipe_flag
                            )

                # Throw an error if dependency not found in any recipe file
                # TODO: should throw an error if dependency not found in one of the recipe files
                if not recipe_flag:
                    if dependency_path.name != "renv.lock":
                       self.warning(
                           "Dependency file %s/%s not used in any %s recipe files %s.",
                           tool_name,
                           dependency_path.name,
                           recipe_type,
                           tree.get("path").relative_to(self.src_path),
                       )

    def _check_env_dir(self, env_tree: dict):
        """

        Args:
            env_tree:

        Returns:

        """
        envs_found = []
        envs_sourced = []
        for env_path in env_tree.get("current_files", []):
            # Skip if not env file
            if env_path.suffix != ".env":
                continue
            envs_found += [env_path]
            # Check if basename of env file is present in label list
            if env_path.stem not in self.labels_all:
                self.warning(
                    "Environment file %s does not correspond to any process label.",
                    env_path.name,
                )
            # Check if this file has been sourced in main.nf (script score in
            # processes_from_workflow)
            for process, process_scope in self.processes_from_workflow.items():
                source_flag = False
                # If basename of env file correspond to one of the labels used in process
                if env_path.stem in process_scope.get("label", []):
                    # If there is a script scope in the process
                    if script := process_scope.get("script", []):
                        for line in script:
                            if re.search(
                                fr"(source|\.)*{env_path.relative_to(self.src_path)}",
                                line,
                            ):
                                source_flag = True
                                envs_sourced += [env_path]
                    # If env file not sourced in the actual process
                    if not source_flag:
                        self.warning(
                            "Process %s with label %s does not source %s.",
                            process,
                            env_path.stem,
                            env_path.relative_to(self.src_path),
                        )

        if envs_unsourced := set(envs_found) - set(envs_sourced):
            for env_path in sorted(envs_unsourced):
                self.warning(
                    "Env file %s not used in the workflow.",
                    env_path.relative_to(self.src_path),
                )

    def get_labels_from_folders(self):
        """Parse information from recipes and modules folders

        Returns:
            labels_from_folders(list): list of tools related to modules, conda, singularity and
            docker files
        """
        geniac_dirs = OrderedDict(
            (
                geniac_dir,
                {
                    "tree": self.project_tree.get(
                        self.default_config.get(GeniacLint.GENIAC_DIRS, geniac_dir)
                    ),
                    "get_labels": getattr(
                        self, f"_get_labels_from_{geniac_dir}_dir", None
                    ),
                    "check_dir": getattr(self, f"_check_{geniac_dir}_dir", None),
                },
            )
            for geniac_dir in self.default_config.options(GeniacLint.GENIAC_DIRS)
        )

        geniac_trees = OrderedDict(
            (f"{geniac_dir}_tree", geniac_scope.get("tree", {}))
            for geniac_dir, geniac_scope in geniac_dirs.items()
        )

        # Get labels first
        for _, geniac_dir in geniac_dirs.items():
            if geniac_dirpath := geniac_dir.get("tree", {}).get("path"):
                if not geniac_dirpath.exists():
                    continue
            if get_label := geniac_dir.get("get_labels"):
                self.labels_from_folders |= get_label(
                    **{
                        arg: geniac_trees.get(arg)
                        for arg in getfullargspec(get_label).args
                        if arg != "self"
                    }
                )

        # Then check directories
        for _, geniac_dir in geniac_dirs.items():
            if geniac_dirpath := geniac_dir.get("tree", {}).get("path"):
                if not geniac_dirpath.exists():
                    continue
            if check_dir := geniac_dir.get("check_dir"):
                check_dir(
                    **{
                        arg: geniac_trees.get(arg)
                        for arg in getfullargspec(check_dir).args
                        if arg != "self"
                    }
                )

        # Check if singularity and docker have the same labels
        if container_diff := sorted(
            list(
                set(
                    self.labels_from_folders.get("singularity", [])
                ).symmetric_difference(set(self.labels_from_folders.get("docker", [])))
            )
        ):
            self.warning(
                "Some recipes are missing either in docker or singularity folder %s.",
                container_diff,
            )

        return self.labels_from_folders

    def check_labels(
        self,
    ):
        """Check labels"""

        # Check if there is any inconsistencies with labels in other parts of config files (post,
        # envCustom) in global nxf_config scope
        for extra_section in (
            "params.geniac.containers.cmd.post",
            "params.geniac.containers.cmd.envCustom",
        ):
            self.nxf_config_container.check_labels_in_section(
                extra_section, self.labels_all
            )

        # List of candidate tools defined in a process
        # using a label defined by a variable
        listCandidateToolsWitlLabelVariable = []
        for process, process_scope in self.processes_from_workflow.items():
            # Get the diff of process labels not present in process scope in config
            # files and present within geniac tools scope
            if process_scope.get("label") == None:
                matched_labels = []
                unmatched_labels = []
            else:
                matched_labels = [
                    label
                    for label in process_scope.get("label")
                    if label not in self.labels_from_process_config
                    and label in self.labels_all
                ]
                unmatched_labels = [
                    label
                    for label in process_scope.get("label")
                    if label not in self.labels_all
                    and label not in self.labels_from_process_config
                ]

            if len(matched_labels) == 0:
                if process_scope.get("labelVariable") != None:
                    if len(process_scope.get("labelVariable")) > 0:
                        foundLabelWithPrefix = False
                        candidateTools = []
                        for labelVariable in process_scope.get("labelVariable"):
                            startWithLabel = list(filter(re.compile(r"^"+labelVariable+r".*").match, self.labels_all))
                            if len(startWithLabel) > 0:
                                foundLabelWithPrefix = True
                                candidateTools = startWithLabel
                                listCandidateToolsWitlLabelVariable = listCandidateToolsWitlLabelVariable + candidateTools
                        if foundLabelWithPrefix:
                            self.info(
                                    "The process '%s' has a label defined by a variable. Candidate tool(s): '%s'.",
                                    process,
                                    candidateTools
                                    )
                        else:
                            self.error(
                                    "The process '%s' has a label defined by the variable '%s' and associated prefix label '%s' but no candidate tool has been found with such a prefix. Check the nextflow process to be sure that the syntax 'label (params.someValue ?: 'toolPrefix')' has been used. 'toolPrefix' must match one of the available labels among the following list:\n'%s'.",
                                    process,
                                    process_scope.get("labelVariableParams"),
                                    process_scope.get("labelVariable"),
                                    self.labels_all
                            )
                    else:
                        self.error(
                                "The process '%s' does not have any label which corresponds to a tool. Available labels are:\n'%s'.",
                                process,
                                self.labels_all
                        )
                else:
                    self.error(
                            "The process '%s' does not have any label which corresponds to a tool. Available labels are:\n'%s'.",
                            process,
                            self.labels_all
                    )

            if len(matched_labels) > 1:
                self.error(
                    "Use only one recipes, modules or geniac.tools label for the process %s %s. "
                    "A process should have only one geniac.tools label.",
                    process,
                    matched_labels,
                )

            if len(unmatched_labels) >= 1:
                process_path = self.get_config_path(
                    GeniacLint.GENIAC_CHECK_CONFIG, "process", single_path=True
                ).relative_to(self.src_path)
                self.error(
                    "Label(s) %s from process %s in the file %s not defined in the file %s.",
                    unmatched_labels,
                    process,
                    Path(process_scope.get("NextflowScriptPath")).relative_to(
                        self.src_path
                    ),
                    process_path,
                )

        # Get the difference with labels from geniac tools and folders and labels used
        # in the workflow
        cross_labels = [
            label
            for label in self.labels_all
            if label not in self.labels_from_workflow and label != "onlyLinux"
        ]
        if len(cross_labels) >= 1:
            if len(listCandidateToolsWitlLabelVariable) >=1:
                notFound = set(cross_labels) - set(listCandidateToolsWitlLabelVariable)
                if len(notFound) >= 1:
                    self.warning(
                        "You have recipes, modules or geniac.tools label(s) that are not used in workflow "
                        "scripts %s.",
                        notFound,
                    )
            else:
                self.warning(
                    "You have recipes, modules or geniac.tools label(s) that are not used in workflow "
                    "scripts %s.",
                    cross_labels,
                )

    def check_labels_containers(
        self, container
    ):
        """Check labels for containers"""
        for label_name in ["modules"]:
            if container_diff := sorted(
                list(
                    set(
                        self.labels_from_folders.get(container, [])
                    ).intersection(set(self.labels_from_folders.get(label_name, [])))
                )
            ):
                self.error(
                        "Some %s recipes are also used by the %s labels: %s. This probably means that you have added %s recipes generated by geniac in the source code repository, then delete these recipes from your source code.",
                    container, label_name, container_diff, container,
                )

        for label_name in ["geniac"]:
            if container_diff := sorted(
                list(
                    set(
                        self.labels_from_folders.get(container, [])
                    ).intersection(set(self.labels_from_configs.get(label_name, [])))
                )
            ):
                self.error(
                        "Some %s recipes are also used by the %s labels: %s. This probably means that you have added %s recipes generated by geniac in the source code repository, then delete these recipes from your source code.",
                    container, label_name, container_diff, container,
                )

    def check_extra_section_geniac_config(self):
        labels = set(self.labels_from_folders.get("modules", []) + self.labels_from_configs.get("geniac", []))
        for extra_section in (
            "params.geniac.containers.yum",
            "params.geniac.containers.git",
        ):
            self.nxf_config_container.check_labels_in_section(
                extra_section, labels
            )

    def check_labels_renv(
        self
    ):
        """Check labels for renv"""

        for (folder_name, label_list) in self.labels_from_folders.items():
            if folder_name != 'conda':
                for label_name in label_list:
                    if bool(re.match(r"^renv.*", label_name)):
                        self.error("In the folder for '%s', you have the label '%s'. Label which starts by 'renv' is only allowed for tools with R and renv. Change the name of your label.",
                                folder_name,
                                label_name)

        for (config_name, label_list) in self.labels_from_configs.items():
            if config_name == 'geniac':
                for label_name in label_list:
                    if bool(re.match("^renv.*", label_name)):
                        self.error("In the config for '%s', you have the label '%s'. Label which starts by 'renv' is only allowed for tools with R and renv. Change the name of your label.",
                                config_name,
                                label_name)



    def check_labels_conda_geniac(
        self
    ):
        """Check that a conda recipe has its label defined in geniac.config"""

        labels = self.labels_from_folders.get('conda', [])
        labels_in_geniac = self.labels_from_configs.get("geniac", [])
        labels_in_workflow = self.labels_from_workflow
        for label in labels:
            if not bool(re.match("^renv.*", label)):
                if label in labels_in_workflow:
                    if label in labels_in_geniac:
                        self.debug("The conda recipe corresponding to the label '%s' is declared in the geniac.config file.", label)
                    else:
                        self.error("The conda recipe corresponding to the label '%s' has not been declared in the geniac.config file. It should be added in the section params.geniac.tools", label)


    def run(self):
        """Execute the main routine

        Returns:

        """


        # Check directory and setup directory flags
        self.check_tree_folder()

        # Get list of labels from main nextflow script
        self.get_processes_from_workflow()

        # Get list of labels from project.config and geniac.config files
        self.get_labels_from_config_files()

        # Get labels from folders
        self.get_labels_from_folders()

        # Check if there is any inconsistency between the labels from configuration
        # files and the main script
        self.check_labels()

        # Check if there is any inconsistency between the labels  in the extra sections
        # from geniac.config
        self.check_extra_section_geniac_config()

        # Check that labels from container recipes have not been used elsewhere.
        # This checks that the containers receipes have not been pushed in the git repository
        self.check_labels_containers(container="singularity")
        self.check_labels_containers(container="docker")

        # Check labels for renv
        self.check_labels_renv()

        # Check that a conda recipe has its label defined in geniac.config"""
        self.check_labels_conda_geniac()

        # End the run with exit code
        if self.error_flag:
            raise SystemExit(1)


def find_duplicates(listOfElem):
    """Extract duplicates in a list"""
    if len(listOfElem) != len(set(listOfElem)):
        return list(set([x for x in listOfElem if listOfElem.count(x) > 1]))
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""options.py: Geniac CMake option class"""

import logging
import re
import shutil
import os
from pathlib import Path

from geniac.cli.commands.init import GeniacInit

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacClean(GeniacInit):
    """Geniac CMake Recipes class"""

    def __init__(
        self,
        *args,
        src_path: str = None,
        **kwargs,
    ):
        """Init flags specific to GRecipes command"""
        super().__init__(
            *args, src_path=src_path, post_clean=True, init_build=False, **kwargs
        )

    def clean_build(self):
        """Clean build directory"""

        build_dir = self.working_dirs["build"].as_posix()
        if os.path.isdir(build_dir):
            shutil.rmtree(build_dir)
            os.mkdir(build_dir)
            self.info("The folder '%s' has been cleaned.", build_dir)

    def run(self):
        """

        Returns:

        """
        self.clean_build()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""config.py: Nextflow configuration file parser"""

from pathlib import Path
import re
import typing
from collections import ChainMap, OrderedDict

from geniac.cli.parsers.base import DEFAULT_ENCODING, GeniacBase, GeniacParser, PathLike

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


class NextflowConfig(GeniacParser):
    """Nextflow config file parser"""

    # Param = value
    PARAM_RE = re.compile(
        r"^ *(?P<scope>[\w.]+(?=\.))?\.?(((?P<property>[\w]+)\s*=\s*"
        r"(?P<elvis>[.\w]+\s*\?:\s*)?(?P<value>([\"\']?.*[\"\']?)|(\d+\.?\w*)|"
        r"(\[[\w\s\'\"/,-]*])|({[\w\s\'\"/,.\-*()]*})))|"
        r"((?P<includeConfig>includeConfig) +['\"](?P<confPath>[\w/.]+))['\"].*) *$"
    )
    SCOPE_RE = re.compile(
        r"^ *(['\"]?(?P<scope>[\w]+)(?<!try)['\"]?|"
        r"(?P<selector>[\w]+) *: *(?P<label>[\w]+)|"
        r"(?P<beforeClose>})?(?P<other>.+)(?<!\$)) *{ *(?P<afterClose>})?$"
    )
    ESCOPE_RE = re.compile(r"^ *}\s*$")
    
    SINGLE_LINE_WITH_LABEL_SCOPE_RE = re.compile(r"^\s*(?P<singleLine>withLabel\s*:\s*\w+)\s*{(?P<singleLineInBracket>.*)}\s*$")


    def _check_config_scope_format(self, nxf_config_scope: str, scope: dict):
        """Check all the properties in a Nextflow config scope if there is a pattern related to the
        format.

        Args:
            nxf_config_scope (str): Name of the Nextflow config scope
            scope (dict): related scope with the nested Nextflow parameters
        """
        properties_pattern = [
            re.compile(pattern)
            for pattern in self.get_config_scope_option_list(
                nxf_config_scope, "patterns"
            )
        ]
        if properties_pattern:
            for property_name, [property_value] in scope.items():
                is_formatted = False
                for reg in properties_pattern:
                    if reg.search(property_value):
                        is_formatted = True
                        break
                if not is_formatted:
                    self.warning(
                        "Parameter %s is not correctly formatted. Read the documentation for more "
                        "details.",
                        ".".join((nxf_config_scope, property_name)),
                    )

    def _check_config_scope(self, nxf_config_scope: str, scope: dict):
        """Do all the checks related to a specific Nextflow config scope

        Args:
            nxf_config_scope (str): Name of the Nextflow config scope
            scope (dict): related scope with the nested Nextflow parameters
        """
        # If there is a pattern related to the format of the scope, we check all the properties
        self._check_config_scope_format(nxf_config_scope=nxf_config_scope, scope=scope)

        config_properties = [
            *self.get_config_scope_option_list(nxf_config_scope, "paths"),
            *self.get_config_scope_option_list(nxf_config_scope, "properties"),
        ]
        # If there is mandatory properties, we check against default values and/or prohibited
        # patterns
        for config_prop in config_properties:
            default_values = self.get_config_section_items(
                f"scope.{nxf_config_scope}.values.default"
            ).get(config_prop, [])
            cfg_vals = scope.get(config_prop, [None])

            for cfg_val in cfg_vals:
                # If the values doesn't match default values
                if (
                    config_prop
                    and default_values
                    and cfg_val not in [_.strip("'\"") for _ in default_values]
                ):

                    if cfg_val is not None:
                        self.warning(
                            'Value "%s" of %s.%s parameter'
                            " in file %s doesn't correspond to one of the expected values [%s].",
                            cfg_val,
                            nxf_config_scope,
                            config_prop,
                            self.path.relative_to(self.src_path),
                            ", ".join(
                                [
                                    _ if '"' in _ or "'" in _ else f"'{_}'"
                                    for _ in filter(None, default_values)
                                ]
                            ),
                        )
                    else:
                        self.error(
                            "Missing %s.%s parameter in file %s.",
                            nxf_config_scope,
                            config_prop,
                            self.path.relative_to(self.src_path),
                        )

    def check_config_scope(self, nxf_config_scope: str, skip_nested_scopes=None):
        """Check if the given scope is in an NextflowConfig instance

        Args:
            nxf_config_scope (str): Scope checked in the Nextflow configuration
            skip_nested_scopes:
        """
        self.info(
            "Checking %s scope in %s.",
            nxf_config_scope,
            self.path.relative_to(self.src_path),
        )

        skip_nested_scopes = [""] if skip_nested_scopes is None else skip_nested_scopes

        scope = self.get(nxf_config_scope, "missing")
        # If scope is empty and required
        if (
            nxf_config_scope
            and scope == "missing"
            and self.default_config.getboolean(f"scope.{nxf_config_scope}", "required")
        ):
            self.error(
                "Required section %s in Nextflow configuration file %s is missing.",
                nxf_config_scope,
                self.path.relative_to(self.src_path),
            )
        # Else if the scope is empty
        elif nxf_config_scope and not scope:
            self.error(
                "Section %s in Nextflow configuration file %s is empty.",
                nxf_config_scope,
                self.path.relative_to(self.src_path),
            )

        # Check if config_paths/config_props in the Nextflow config corresponds to
        # their default values and are not prohibited (if a prohibited pattern is available)
        if scope != "missing":
            self._check_config_scope(nxf_config_scope, scope)

        # Recursive call to do checks on nested scopes
        for nested_scope in self.get_config_scope_option_list(
            nxf_config_scope, "scopes"
        ):
            if nested_scope not in skip_nested_scopes:
                self.check_config_scope(".".join((nxf_config_scope, nested_scope)))

    @staticmethod
    def check_labels_in_section(nxf_config, section_name, labels):
        """Check if given labels exists within a section"""
        if x_section := nxf_config.get(section_name):
            # For each label in the section scope
            for label in x_section:
                # If label is not present in labels
                if label not in labels:
                    nxf_config.error(
                        "Label %s of %s is not defined neither in params.geniac.tools nor availabe as a module.",
                        label,
                        section_name,
                    )

    def _set_scope(self, values: dict, scope_idx: str, def_flag: bool):
        """Set a new scope according to the actual scope_idx

        Args:
            values (dict): group dict from matching pattern
            scope_idx (str): Index of the last scope in content tree structure
            def_flag (bool): Flag corresponding to definition groovy blocks

        Returns:
            scope_idx (str): Index of the last scope in content tree structure
            selector (str): Setup if there is any with* nextflow selector
            def_flag (bool): Flag corresponding to definition groovy blocks
        """
        # If scope add it to the scopes dict
        if scope := values.get("scope"):
            scope_idx = scope if not scope_idx else ".".join((scope_idx, scope))
        # If there is also a selector on the line add them to scope_idx
        if (selector := values.get("selector")) and (label := values.get("label")):
            scope_idx = (
                ".".join((selector, label))
                if not scope_idx
                else ".".join((scope_idx, selector, label))
                if selector not in scope_idx
                else ".".join((scope_idx, label))
            )
        # If close pattern, remove the last scope_idx
        if values.get("beforeClose"):
            scope_idx = ".".join(scope_idx.split(".").pop()) if "." in scope_idx else ""
        # Add the rest of the line in other section
        if scope := values.get("other"):
            def_flag = True if "def" in scope else def_flag
            scope_idx = (
                "other"
                if not scope_idx or scope_idx == "other"
                else ".".join((scope_idx, "other"))
            )
        # Init the scope with the scope_idx only if it does not exist
        if scope_idx not in self.content:
            self.content[scope_idx] = OrderedDict()
        # Update scope_idx if there is a closing bracket at the end of the match
        if values.get("afterClose"):
            scope_idx = ".".join(scope_idx.split(".").pop())
        return scope_idx, selector, def_flag

    def _set_param(
        self, values: dict, scope_idx: str, line_idx: int, warnings: bool = True
    ):
        """Set Nextflow parameter in content property

        Args:
            values (dict): group dict from matching pattern
            scope_idx (str): Index of the scope in content tree structure
            line_idx (int): Index of the current line in the file
        """
        prop_key = "property" if values.get("property") else "includeConfig"
        value_key = "value" if values.get("value") else "confPath"
        prop = values.get(prop_key, "")
        value = values.get(value_key, "")
        param_list = list(filter(None, (scope_idx, values.get("scope"), prop)))
        param_idx = ".".join(param_list) if len(param_list) > 1 else param_list[0]
        self.debug(
            "FOUND property %s with value %s in scope %s.",
            values.get(prop_key),
            values.get(value_key),
            param_idx,
        )

        # Issue #278: Geniac lint: error due to comments in the manifest section of nextflow.config
        if value == None:
            value = ""

        value = (
            value.strip('"')
            if '"' in value
            else value.strip("'")
            if "'" in value
            else value
        )
        # If parameter has already been defined in a previous configuration
        # file
        if (
            value
            and warnings
            and param_idx in self.content
            and self.content[param_idx]
            and prop_key != "includeConfig"
        ):
            history_paths = [
                conf_path.relative_to(self.src_path).name
                for conf_path in self.loaded_paths
            ]
            extra_msg = (
                f" in a previous configuration file {history_paths}"
                if self.loaded_paths
                else " in the same file"
            )
            self.warning(
                "Parameter %s from %s at line %s has already been defined%s.",
                param_idx,
                self.path.relative_to(self.src_path),
                line_idx + 1,
                extra_msg,
            )
        self.content[param_idx] = (
            [value]
            if param_idx not in self.content
            else self.content[param_idx] + [value]
        )

    def _read(
        self,
        in_file: typing.Union[typing.IO, typing.BinaryIO],
        encoding: str = DEFAULT_ENCODING,
        in_path: PathLike = None,
        flush_content: bool = False,
        warnings: bool = True,
        **kwargs,
    ):
        """Load a Nextflow config file into content property

        Args:
            in_file (BinaryIO): input Nextflow config file
            encoding (str): encoding type used to read the input file
            in_path (PathLike): path to input file
            flush_content (bool): flag used to flush previous content before reading
            warnings (bool): flag to turn on/off warning messages

        Returns:
            content (
        """
        def_flag = False
        selector = None
        scope_idx = ""
        content_cache = self.content
        for line_idx, line in enumerate(
            super()._read(in_file, encoding=encoding, flush_content=True, **kwargs)
        ):
            # Single line scope (i.e. open and close brackets on the same line in process.config)
            if match := self.SINGLE_LINE_WITH_LABEL_SCOPE_RE.match(line):
                if Path(in_path).relative_to(self.src_path).as_posix() == "conf/process.config":
                    self.error("In '%s' file, the 'withLabel' process selector if formatted on a single line:\n\n%s\nFor efficient parsing, modify the code using multi-line format:\n\n%s\n%s\n%s\n",
                            Path(in_path).relative_to(self.src_path),
                            line,
                            " " + match.groupdict().get("singleLine") + " {",
                            " " + match.groupdict().get("singleLineInBracket"),
                            " }"
                            )
            # Pop scope index list if we find a curly bracket
            # Turn off def flag if we reach the last scope in a def
            if self.ESCOPE_RE.match(line):
                depth = 1 if not selector else 2
                scope_idx = ".".join(scope_idx.split(".")[:-depth])
                selector = None
                if not scope_idx and def_flag:
                    def_flag = False
                continue
            # If we find a new scope
            if match := self.SCOPE_RE.match(line):
                (scope_idx, selector, def_flag) = self._set_scope(
                    match.groupdict(), scope_idx, def_flag
                )
                continue
            # If we are not in a def scope and we find a parameter
            if not def_flag and (match := self.PARAM_RE.match(line)):
                self._set_param(
                    match.groupdict(), scope_idx, line_idx, warnings=warnings
                )
                continue
        content = self.content.copy()
        # If we don't want to flush, then we merge with previous content
        if not flush_content:
            self.content.update(content_cache)
        return content


class NextflowConfigContainer(GeniacBase, ChainMap):
    """Container object to save nextflow config files"""

    def append(self, item: NextflowConfig):
        """Add a new Nextflow Config"""
        self.maps.append(item.content)

    def check_labels_in_section(self, section_name, labels):
        """Check if given labels exists within a section"""
        return NextflowConfig.check_labels_in_section(self, section_name, labels)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_base.py: Test geniac.base module"""

from configparser import ConfigParser

import pytest

from geniac.cli.utils.base import GeniacBase, glob_solver

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


@pytest.fixture
def default_gbase():
    """Define a GCommand instance for testing"""
    return GeniacBase()


def test_glob_solver(shared_datadir):
    """"""
    assert glob_solver(input_path=str(shared_datadir / "modules" / "*.txt")) == []
    assert glob_solver(
        input_path=str(shared_datadir / "modules" / "fromSource" / "*.txt")
    ) == [shared_datadir / "modules" / "fromSource" / "CMakeLists.txt"]
    assert glob_solver(
        input_path=str(shared_datadir / "modules" / "**" / "*.txt")
    ) == sorted(
        [
            shared_datadir / "modules" / "fromSource" / "CMakeLists.txt",
            shared_datadir / "modules" / "fromSource" / "helloWorld" / "CMakeLists.txt",
            shared_datadir / "modules" / "fromSource" / "dolorSit" / "CMakeLists.txt",
        ]
    )
    assert glob_solver(
        input_path=str(shared_datadir / "modules" / "**/hello*" / "*.txt")
    ) == sorted(
        [
            (
                shared_datadir
                / "modules"
                / "fromSource"
                / "helloWorld"
                / "CMakeLists.txt"
            ),
        ]
    )


def test_gbase(default_gbase):
    """Check if default GBase has been instantiated correctly with all default GBase properties"""
    assert default_gbase.config_file is None
    assert isinstance(default_gbase.default_config, ConfigParser)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Foobar.py: Description of what foobar does."""

# import pytest

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


def test_main():
    """Check if we can instantiate a GConfor object with a folder"""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""options.py: Geniac CMake option class"""

import logging
import re

from geniac.cli.commands.init import GeniacInit

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacOptions(GeniacInit):
    """Geniac CMake option class"""

    def __init__(
        self,
        *args,
        src_path: str = None,
        **kwargs,
    ):
        """Init flags specific to GOptions command"""
        super().__init__(
            *args, src_path=src_path, post_clean=True, init_build=True, **kwargs
        )

    def _format_cmake_help(self, cmake_output: list) -> list:
        """Format CMake -LAH output as a comprehensible list"""
        formatted_output = []
        cmake_reg = re.compile(
            self.default_config.get("geniac.install", "cmakeOutputPattern")
        )
        for line in cmake_output:
            if match := cmake_reg.match(line):
                groups = match.groupdict()
                formatted_output += [
                    (
                        groups.get("option_name"),
                        groups.get("option_type"),
                        groups.get("option_value"),
                    )
                ]
        return formatted_output

    def cmake_help(self, format_output: bool = False) -> list:
        """Return help formatted from cmake -LAH command"""

        cmake_lah = (
            self._subprocess_run(
                ["cmake", "-LAH", (self.working_dirs["src"] / "geniac").as_posix()],
                check=True,
                cwd=self.working_dirs["build"],
                cmd_name="CMake -LAH",
            )
            if self.working_dirs.get("src")
            else None
        )

        return (
            self._format_cmake_help([_ for _ in cmake_lah.stdout.split("\n") if _])
            if cmake_lah and cmake_lah.stdout and format_output
            else [cmake_lah.stdout]
            if cmake_lah and cmake_lah.stdout
            else []
        )

    def run(self):
        """

        Returns:

        """
        cmake_help = self.cmake_help(format_output=True)
        # Print in the console cmake_help if
        if self.logger.root.level >= logging.WARNING:
            for _ in cmake_help:
                print(f"{_[0]}:{_[1]}='{_[2]}'")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test_check.py: Test geniac.check module"""

from pathlib import Path

import pytest

from geniac.cli.commands.lint import GeniacLint

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


@pytest.fixture
def gcheck_data(shared_datadir):
    """Instantiate GCheck command with shared datadir from pytest-datadir"""
    return GeniacLint(shared_datadir)


def test_data_gcheck(gcheck_data, shared_datadir):
    """Check if  GChek with data has been instantiated correctly"""
    assert gcheck_data.src_path == Path(shared_datadir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""base.py: Define GBase class and several utils functions"""

import errno
import logging
import os
import re
import sys
from abc import ABC
from configparser import ConfigParser, ExtendedInterpolation
from distutils.dir_util import copy_tree
from json import loads as json_loads
from os.path import basename, dirname, isfile
from pathlib import Path
from shutil import rmtree
from sys import exit as sys_exit
from tempfile import TemporaryDirectory, mkdtemp
from typing import NamedTuple

import shutil

import validators
from git import Repo

# TODO: use importlib.resources in the future
# https://importlib-resources.readthedocs.io/en/latest/migration.html#pkg-resources-resource-filename
from pkg_resources import resource_filename, resource_stream

from geniac.cli.utils.logging import LogMixin
from geniac import __version__

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2021"

_logger = logging.getLogger(__name__)
CMAKE_OPTION_PREFIX = "cmake_"
# Sadly, Python fails to provide the following magic number for us.
ERROR_INVALID_NAME = 123
"""
Windows-specific error code indicating an invalid pathname.

See Also
----------
https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
    Official listing of all such codes.
"""

### Geniac variables
GENIAC_CONFIG_FILES=['conda.config',
        'multiconda.config',
        'path.config',
        'multipath.config',
        'apptainer.config',
        'docker.config',
        'podman.config',
        'singularity.config',
        'cluster.config']

def load_logging_config(logging_config_path) -> dict:
    """Read logging config file in json format and return it as a dict"""
    with logging_config_path.open("rb") as logging_config_file:
        return json_loads(logging_config_file.read().decode())


def _dir_checker(value: str):
    """
    Exit if the path is not correct
    """
    path = value if isinstance(value, Path) else Path(value)
    if not path and not path.is_dir():
        _logger.critical("Path %s does not exist.", path)
        sys_exit(1)
    return path.resolve()


def glob_solver(input_path: str, lazy_flag: bool = False):
    """
    Use native glob solver to expand glob patterns from an input path

    Args:
        input_path (str): input file path
        lazy_flag (bool): expand to existing paths or generate paths without any check

    Returns:
        output_paths (list): list of expanded paths
    """
    for glob_string in ("**", "*"):
        if glob_string in input_path:
            # Get the tuple of elements with 3 parts (base_dir, glob_pattern, filename)
            input_path_parts = input_path.partition(glob_string)
            (stem_path, file_name) = (
                Path(dirname(input_path_parts[0])).resolve(),
                basename(input_path),
            )
            no_glob_file = "*" not in file_name
            # IF LAZY
            #   IF no_glob_file THEN we make a pattern without the file_name and add it
            #       to the path after the glob expansion
            #   ELSE THEN we make a pattern with the file_name and doesn't add it after the glob
            #       expansion
            # ELSE THEN we make a pattern with the file_name and doesn't add it after the glob
            #   expansion
            glob_pattern = "/".join(
                [
                    path_elt
                    for path_elt in [
                        basename(input_path_parts[0]),
                        input_path_parts[1] + dirname(input_path_parts[2]).rstrip("/")
                        if basename(input_path_parts[2]) == file_name
                        else "" + dirname(input_path_parts[2]).rstrip("/"),
                        "" if lazy_flag and no_glob_file else file_name,
                    ]
                    if path_elt
                ]
            )
            return (
                [
                    path / Path(file_name)
                    for path in stem_path.glob(glob_pattern)
                    if path.is_dir()
                    and not path.resolve().samefile(stem_path.resolve())
                ]
                if lazy_flag and no_glob_file
                else sorted(stem_path.glob(glob_pattern))
            )
    return [""]


def is_pathname_valid(pathname: str) -> bool:
    """
    CF https://stackoverflow.com/a/34102855
    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = (
            os.environ.get("HOMEDRIVE", "C:")
            if sys.platform == "win32"
            else os.path.sep
        )
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?


class MethodRecord(NamedTuple):
    """Default class to fill args and kwargs for a method"""

    args: list
    kwargs: dict


class GeniacBase(ABC, LogMixin):
    """Abstract base class for Geniac commands"""

    DEFAULT_CONFIG = ("geniac", "cli/data/conf/geniac.ini")
    CACHE_NAME = ".geniac"
    SRC_NAME = "src"
    BUILD_NAME = "build"

    def __init__(
        self,
        src_path: str = Path.cwd(),
        config_file: str = None,
        working_dir: str = None,
        pre_clean: bool = False,
        post_clean: bool = True,
        init_work_path: bool = False,
        **kwargs,
    ) -> None:
        """

        Args:
            src_path (str): path to the Nextflow Project Dir
            config_file (str): path to a configuration file (INI format)
            working_dir (str): path to a geniac working directory
            pre_clean (bool): should we clean every build folder before running any cmd ?
            post_clean (bool): should we clean every temporary folder at the end of the execution ?
            init_work_path (bool): should we init a working directory with source files inside ?
        """


        super().__init__()
        self.config_section = (
            re.sub(r"(?<!^)(?=[A-Z])", ".", self.__class__.__name__).lower()
            if not (_config_section := kwargs.get("config_section"))
            else _config_section
        )


        # geniac command line
        self.geniac_cmd = kwargs.get("which")

        # Project path is optional but if it is not a valid url or a valid path, it will correspond
        # to the default config project.metadata.url
        self._src_path = None
        (self.project_type, self.src_is_working_dir) = self._get_src_type(src_path, working_dir)
        self.src_path = src_path

        # Init and load config files
        self.default_config_file = Path(resource_filename(*self.DEFAULT_CONFIG))
        self.config_file = Path(config_file) if config_file else None
        self.default_config = self._load_config(config_file=self.config_file)
        self._update_default_config(**kwargs)

        self._tmp_dir = None

        self.working_dir_is_valid = self._check_working_dir(working_dir)
        # Working dir correspond to tmp dir or working_dir
        self.working_dir = (
            self.src_path.parent.resolve()
            if self.project_type == "wd"
            else Path(working_dir).resolve()
            if working_dir and self.working_dir_is_valid
            else Path().cwd()
        )
        self.working_dirs = {
            "build": self.working_dir / self.BUILD_NAME,
            "src": self.working_dir / self.SRC_NAME,
            "cache": self.working_dir / self.CACHE_NAME,
        }

        if src_path and init_work_path:
            self.init_working_path(
                working_dir,
                post_clean,
                pre_clean,
                kwargs.get(
                    "branch", self.default_config["project.metadata"].get("branch")
                ),
            )

    def _load_config(self, config_file: Path = None):
        """Load default configuration file and update option with config_file

        Returns:
            :obj:`configparser.ConfigParser`: config instance
        """
        config = ConfigParser(
            interpolation=ExtendedInterpolation(), allow_no_value=True
        )

        # Read default config file
        config.optionxform = str
        config.read_string(resource_stream(*self.DEFAULT_CONFIG).read().decode())
        if config_file:
            # Read configuration file
            config.read(config_file)
        return config

    def _update_default_config(self, **kwargs):
        """Update default config according to arguments in kwargs which should come from argparse
        in the constructor"""
        if self.config_section in self.default_config.sections():
            for option in self.default_config.options(self.config_section):
                if option_value := kwargs.get(option):
                    self.default_config.set(self.config_section, option, option_value)

    def _check_working_dir(self, working_dir: str, info: bool = True) -> bool:
        """Check if given working dir is valid"""

        if working_dir and (
            not str(working_dir).startswith("ssh")
            or not str(working_dir).startswith("http")
        ):
            working_path = Path(working_dir)
            cache_path = working_path / self.CACHE_NAME
            is_existing_path = working_path.exists()
            is_valid_path = is_pathname_valid(working_path.as_posix())
            is_empty_dir = working_path.is_dir() and not list(working_path.iterdir())
            has_cache_dir = is_existing_path and cache_path.exists()
            # If empty folder or uninitialized folder or folder with .geniac sub folder
            if (
                (working_dir and is_empty_dir)
                or (is_valid_path and not is_existing_path)
                or has_cache_dir
            ):
                return True
        if info:
            self.info(
                "Working directory %s is not an empty directory, a valid path or doesn't have "
                "expected cache folder %s. A temporary folder will be created instead",
                working_dir or "",
                self.CACHE_NAME,
            )
        return False

    def _get_src_type(self, src_path: [str, None], working_dir: str) -> (str, bool):
        """Check if given input is a valid url or a valid project path"""
        self.debug("Checking if source path %s is a path or an url", src_path)
        is_working_dir = self._check_working_dir(src_path, info=False)

        if src_path and validators.url(str(src_path)) or (src_path and str(src_path).startswith("ssh")) or (src_path and str(src_path).startswith("http")):
            src_type = ("url", False)
        else:
            if is_working_dir:
                src_type = ("wd", is_working_dir)
            else:
                if src_path and Path(src_path).exists():
                    src_type = ("path", False)
                else:
                    src_type = ("default", False)

        if self.geniac_cmd == "init":
            if src_type[0] == "path" and not src_type[1]:
                if working_dir == str(os.getcwd()):
                    self.error(
                            "You have launched 'geniac init' command inside the working directory your want to create: '%s'. You must be located in another directory when using the command 'geniac init'.",
                            working_dir
                            )
                    sys_exit(1)
                if os.path.isdir(working_dir):
                    self.error(
                            "The working directory you want to create '%s' already exists. Use another directory instead.",
                            working_dir
                            )
                    sys_exit(1)


            if src_type[0] == "wd" and src_type[1]:
                if not os.path.isdir(src_path):
                    self.critical("The folder '%s' to the Nextflow repository does not exist.", src_path)
                    sys_exit(1)

        return src_type


    def init_working_tree(self):
        """Initialize working directory with basic tree structure"""
        self.info(
            "Initializing geniac working directory tree at %s",
            self.working_dir.as_posix(),
        )
        if not self.working_dir.exists():
            self.working_dir.mkdir(parents=True)
        for nested_dir_name, nested_dir_path in self.working_dirs.items():
            if not nested_dir_path.exists():
                nested_dir_path.mkdir()
            elif nested_dir_path.is_dir():
                self.warning(
                    "%s folder %s already exists",
                    nested_dir_name.capitalize(),
                    nested_dir_path.as_posix(),
                )
            else:
                self.error(
                    "Path %s should be an empty folder", nested_dir_path.as_posix()
                )

    def init_working_src_folder(self, git_branch: str):
        """Copy or clone the project into src temp folder"""
        # If src_path, copy it into src
        if self.project_type == "path":
            self.info(
                "Copy content of %s into geniac working directory %s",
                self.src_path.as_posix(),
                self.working_dirs["src"].as_posix(),
            )
            copy_tree(self.src_path.as_posix(), self.working_dirs["src"].as_posix())
        # Else try to clone project.metadata.url into src folder
        elif self.project_type in ("url", "default"):
            self.info(
                "Using the URL %s to clone %s branch into geniac working directory %s",
                self.src_path,
                git_branch or "default",
                self.working_dirs["src"].as_posix(),
            )

            try:
                Repo.clone_from(
                    self.src_path,
                    self.working_dirs["src"],
                    branch=git_branch,
                    multi_options=["--recurse-submodules"],
                )
            except Exception as error:
                self.error(
                    "Unable to clone the project into the temporary directory.\n%s",
                    error,
                )
           
        else:
            self.error("Unable to initialize src folder at %s", self.src_path)

        ############################################################################
        ### Delete files generated by geniac if they have been pushed in the src dir
        ############################################################################
        for config_file in GENIAC_CONFIG_FILES:
            config_file_path = self.working_dirs["src"].as_posix() + "/conf/" + config_file
            if os.path.isfile(config_file_path):
                os.remove(config_file_path)
                self.info(f"File {config_file} was present in the repo. It has been deleted such that it can be generated by geniac.")

        ######################################################################
        ### Use src/geniac which corresponds to the same version of geniac CLI
        ######################################################################
        if os.path.isdir(self.working_dirs["src"].as_posix()+'/geniac'):
            shutil.rmtree(self.working_dirs["src"].as_posix()+'/geniac')
            shutil.copytree(os.path.dirname(__file__) + '/../../repo', self.working_dirs["src"].as_posix()+'/geniac')
            self.info(f"The geniac folder has been replaced using the geniac folder from the python package.")
        else:
            shutil.copytree(os.path.dirname(__file__) + '/../../repo', self.working_dirs["src"].as_posix()+'/geniac')
            self.info(f"The geniac directory does not exist. It has been created using the geniac folder from the python package.")

    def init_working_path(
        self, working_dir: str, post_clean: bool, pre_clean: bool, git_branch: str
    ):
        """Init working dir and source dir"""
        # Setup working dir

        # IF working_dir has been given and source path is a valid geniac working directory then
        # stop the program
        if self.src_is_working_dir and self.working_dir_is_valid:
            self.critical(
                "Both source path and working directory are geniac working directory. "
                "If the working directory has been given, the source path should be a "
                "valid Nextflow project"
            )

        self.info("project_type: %s", self.project_type)

        if self.project_type != "wd":
            if not self.working_dir_is_valid:
                # Initialize TMP dir if there is no working directory
                self._tmp_dir = TemporaryDirectory() if post_clean else mkdtemp()

                if not post_clean:
                    self.warning(
                        "Post clean option hasn't been set. The temporary folder '%s' will "
                        "not been cleaned at the end !",
                        self._tmp_dir,
                    )

                self.working_dir = (
                    Path(self._tmp_dir.name).resolve()
                    if post_clean
                    else Path(self._tmp_dir).resolve()
                )
                self.working_dirs = {
                    "build": self.working_dir / self.BUILD_NAME,
                    "src": self.working_dir / self.SRC_NAME,
                    "cache": self.working_dir / self.CACHE_NAME,
                }

            # Create basic tree structure
            self.init_working_tree()

        # Clean build directory if asked
        if pre_clean:
            self.clean_build()

        # Copy or clone the project into src temp folder
        if self.project_type != "wd":
            self.init_working_src_folder(git_branch)

    def clean_build(self):
        """Clean build directory"""
        if self.working_dirs["build"].exists():
            self.info("Cleaning build directory %s", self.working_dirs["build"])
            # TODO: catcher l'erreur
            rmtree(self.working_dirs["build"], ignore_errors=True)
            self.working_dirs["build"].mkdir()

    @property
    def src_path(self):
        """Project Dir path property"""
        return self._src_path

    @src_path.setter
    def src_path(self, value):
        """If value is not a directory, set the project dir to the current directory"""
        self._src_path = (
            (Path(value) / "src").resolve()
            if self.project_type == "wd"
            else _dir_checker(value)
            if self.project_type == "path"
            else value
            if self.project_type == "url"
            else value
        )

    @property
    def config_file(self):
        """Configuration file (optional)

        Returns:

        """
        return self._config_file

    @config_file.setter
    def config_file(self, value):
        """"""
        self._config_file = value if value and isfile(value) else None

    @property
    def default_config(self):
        """ConfigParser property

        Returns:
            :obj:`configparser.ConfigParser`: config instance
        """
        return self._config

    @default_config.setter
    def default_config(self, value):
        """Update base dir with project dir"""
        value.set("tree.base", "path", str(self.src_path))
        self._config = value

    def get_config_path(
        self,
        section: str,
        option_name: str,
        single_path: bool = False,
        lazy_glob: bool = False,
    ):
        """Format config option to list of Path objects or Path object if there is only one path

        Args:
            section (str): name of config section
            option_name (str): name of config option
            single_path (bool): flag to enable the return of single path
            lazy_glob (bool): use lazy glob solver without any check

        Returns:
            config_paths (list, Path)
        """

        option = (
            self.default_config.get(section, option_name).split()
            if self.default_config.get(section, option_name)
            else []
        )
        # Get Path instance for each file in the related configparser option. Glob
        # patterns are unpacked here
        result = sorted(
            [
                Path(in_path) if "*" not in in_path else glob_path
                for in_path in option
                for glob_path in glob_solver(in_path, lazy_flag=lazy_glob)
            ]
        )
        return result[0] if len(result) == 1 and single_path else result

    def get_config_subsection(self, subsection):
        """Filter sections to a uniq sub section"""
        return (
            section
            for section in self.default_config.sections()
            if section.startswith(f"{subsection}.")
        )

    def get_config_scope_option_list(self, section: str, option: str) -> list:
        """
        Format option values related to a specific scope section in geniac.ini configuration
        file as a list

        Args:
            section (str): Name of the section in the configuration file (geniac.ini)
            option: Name of the option within the section

        Returns:
            list
        """
        return (
            list(filter(None, config_option.split("\n")))
            if (config_option := self.default_config.get(f"scope.{section}", option))
            else []
        )

    def get_config_section_items(self, section: str) -> dict:
        """Get section items if they exist or return an empty dic"""
        return {
            key: value.split("\n")
            for key, value in (
                self.default_config.items(section)
                if self.default_config.has_section(section)
                else []
            )
        }

    def __exit__(self, *args):
        """Delete temporary folder if it has been correctly initialized"""
        if self._tmp_dir:
            self._tmp_dir.cleanup()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""Geniac Module"""

from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""base.py: CLI geniac interface"""

import logging
import os
import subprocess
from abc import abstractmethod
from shutil import which

from geniac.cli.utils.base import GeniacBase

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacCommand(GeniacBase):
    """Base geniac command"""

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

    def _subprocess_run(
        self,
        cmd: list,
        cmd_name: str = "",
        error_msg: str = "",
        debug_msg: str = "",
        **kwargs
    ):
        """Call subprocess run and log output"""
        cmd_out = None
        cmd = [which(cmd[0])] + cmd[1:]
        working_directory = kwargs.get("cwd", os.getcwd())
        self.info(
            "Running '%s' command:\n%s\n> %s",
            cmd_name,
            working_directory,
            " ".join(cmd),
        )
        try:
            cmd_out = subprocess.run(
                cmd,
                capture_output=kwargs.get("capture_output", True),
                check=kwargs.get("check", True),
                cwd=working_directory,
                encoding="utf8",
            )
        except subprocess.CalledProcessError as error:
            if error and error.stdout:
                self.info(
                    "%sCommand%s standard output:\n%s",
                    debug_msg + "\n" if debug_msg else "",
                    " " + cmd_name if cmd_name else "",
                    str(error.stdout).rstrip(),
                )
            self.critical(
                "%sCommand%s returned non-zero exit status:\n%s",
                error_msg + "\n" if error_msg else "",
                " " + cmd_name if cmd_name else "",
                str(error.stderr).rstrip(),
            )
        else:
            if cmd_out and cmd_out.stdout:
                self.info(
                    "%sCommand%s standard output:\n%s",
                    debug_msg + "\n" if debug_msg else "",
                    " " + cmd_name if cmd_name else "",
                    str(cmd_out.stdout).rstrip(),
                )
            if cmd_out and cmd_out.stderr:
                self.warning(
                    "Command%s standard error:\n%s",
                    " " + cmd_name if cmd_name else "",
                    str(cmd_out.stderr).rstrip(),
                )
        return cmd_out

    @abstractmethod
    def run(self):
        """Main entry point for every geniac sub command"""
        raise NotImplementedError("This class should implement a basic run command")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""base.py: Geniac base file parser"""

import logging
import re
import tempfile
import typing
from abc import abstractmethod
from collections import OrderedDict
from io import StringIO
from json import dumps
from os import PathLike
from pathlib import Path

from dotty_dict import Dotty

from geniac.cli.utils.base import GeniacBase

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)
DEFAULT_ENCODING = "UTF-8"


class GDotty(Dotty):
    """Add merge feature to dotty dict"""

    def update(self, other: dict):
        """Merge with another dotty dict"""
        self._data.update(other)


class GeniacParser(GeniacBase):
    """Geniac file parser"""

    COM_RE = re.compile(
        r"(?P<tdquote>\"{3}[\S\s]*?\"{3})|"
        r"(?P<tquote>\'{3}[\S\s]*?\'{3})|"
        r"(?P<squote>\'[^\']*\')|"
        r"(?P<dquote>\"[^\"]*?\")|"
        r"(?P<scom>//.*?$)|"
        r"(?P<mcom>/\*([\s\S]*?)\*/)",
        re.MULTILINE,
    )

    def __init__(self, *args, **kwargs):
        """Constructor for GParser"""
        super().__init__(*args, **kwargs)
        self.params = None
        self._path = ""
        self._loaded_paths = []
        self._content = OrderedDict()

    @property
    def content(self):
        """Content loaded from input file with read method"""
        return self._content

    @content.setter
    def content(self, value: dict):
        """Content loaded from input file with read method"""
        self._content = (
            value if isinstance(value, GDotty) else GDotty({} if not value else value)
        )

    @property
    def path(self):
        """Content loaded from input file with read method"""
        return self._path

    @path.setter
    def path(self, value):
        """Content loaded from input file with read method"""
        self._path = value

    @property
    def loaded_paths(self):
        """Content(s) loaded from input file with read method"""
        return self._loaded_paths

    @loaded_paths.setter
    def loaded_paths(self, value):
        """Content loaded from input file with read method"""
        self._loaded_paths = value

    def __getitem__(self, item):
        """Get a content item"""
        return self._content[item]

    def __setitem__(self, key, value):
        """Set an item in content"""
        self._content[key] = value

    def __repr__(self):
        """List only values in content dict"""
        return repr(self.content)

    def __contains__(self, item):
        """Check if item is in content dict"""
        return item in self._content

    def __delitem__(self, key):
        """Remove a key from content dict"""
        del self._content[key]

    def get(self, key, default=None):
        """Get method with default option"""
        if key in self.content:
            return self[key]
        return default

    def _remove_comments(self, in_file, temp_file):
        # Remove comments for the analysis

        def match_comments(match):
            """Filter comments from match object"""
            return "" if match.group("mcom") or match.group("scom") else match.group(0)

        input_content = self.COM_RE.sub(match_comments, in_file.read())
        temp_file.write(bytes(input_content, encoding=DEFAULT_ENCODING))
        temp_file.seek(0)
        return temp_file

    @abstractmethod
    def _read(
        self,
        in_file: typing.Union[typing.IO, typing.BinaryIO],
        in_path: PathLike = Path(""),
        **kwargs,
    ):
        """Load a file into content property

        Args:
            in_file (TextIO): input file
            encoding (str): encoding type used to read the input file
            in_path (PathLike): path to input file
            flush_content (bool): flag used to flush previous content before reading
            warnings (bool): flag to turn on/off warning messages

        Returns:
            content (TextIO): Raw content of the file
        """
        self.debug(f"Reading file {in_path}")
        if kwargs.get("flush_content"):
            self.content = OrderedDict()
        return StringIO(in_file.read().decode(kwargs.get("encoding", DEFAULT_ENCODING)))

    def read(
        self,
        in_paths: [str, PathLike, list],
        encoding: str = DEFAULT_ENCODING,
        **kwargs,
    ) -> [bool]:
        """Read and parse a file or an iterable of files

        Args:
            in_paths: path to input file(s)
            encoding (str): name of the encoding used to decode files

        Returns:
            read_ok (list): list of successfully read files
        """
        self.path = in_paths
        if isinstance(in_paths, (str, bytes, PathLike)):
            in_paths = [in_paths]
        read_ok = []
        for in_path in in_paths:
            in_path = Path(in_path) if not isinstance(in_path, PathLike) else in_path
            try:
                with in_path.open(
                    mode="r", encoding=encoding
                ) as input_file, tempfile.TemporaryFile() as temp_file:
                    # Format files before reading
                    temp_file = self._remove_comments(input_file, temp_file)
                    temp_content = self._read(
                        temp_file, encoding=encoding, in_path=in_path, **kwargs
                    )
                    self.debug(
                        "LOADED %s scope:\n%s.",
                        temp_file,
                        dumps(dict(self.content), indent=2),
                    )
                    self.loaded_paths += [in_path]
            except OSError:
                continue
            read_ok.append((in_path, temp_content))
        return read_ok



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""install.py: Geniac init command"""

import logging
import os
from pathlib import Path
import subprocess

from geniac.cli.commands.base import GeniacCommand

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2021"

_logger = logging.getLogger(__name__)


class GeniacInit(GeniacCommand):
    """Geniac working directory initialization"""

    def __init__(
        self,
        *args,
        src_path: str = None,
        init_build: bool = False,
        pre_clean: bool = True,
        post_clean: bool = False,
        **kwargs
    ):
        """Init flags specific to GCheck command"""
        self.post_clean = post_clean
        self.pre_clean = pre_clean

        super().__init__(*args, src_path=src_path, init_work_path=True, **kwargs)

        # First cmake call into the build folder
        if init_build:
            self.init_build_folder()

    def init_build_folder(self):
        """Call cmake without options into the build folder"""
        self._subprocess_run(
            ["cmake", (self.working_dirs["src"] / "geniac").as_posix()],
            check=True,
            cwd=self.working_dirs["build"],
            cmd_name="CMake init",
        )

    def create_sub_shell(self):
        """Create a sub shell with the current conda env if it exists"""
        conda_prefix = os.environ.get("CONDA_PREFIX")
        # Force conda to activate the actual env with default_env var
        shell_env = os.environ | {
            "CONDA_DEFAULT_ENV": conda_prefix.split("/")[-1] if conda_prefix else ""
        }
        conda_env = Path(conda_prefix).name if conda_prefix else ""
        shell_env["CONDA_DEFAULT_ENV"] = conda_env or ""
        self.info(
            "Opening a sub shell with working directory set to %s",
            self.working_dir.resolve(),
        )
        
        self._subprocess_run(
            [os.environ.get("SHELL", os.environ.get("COMSPEC", "sh")), "cd"],
            cwd=self.working_dir.resolve(),
            env=shell_env,
            capture_output=False,
        )

    def run(self):
        """Entry point for init command"""
        if not self.post_clean:
            self.info("geniac init completed")

    def __exit__(self, *args):
        """Delete temporary folder if it has been correctly initialized and post_clean flag"""
        if self._tmp_dir and self.post_clean:
            self._tmp_dir.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""options.py: Geniac CMake option class"""

import logging
import re
import shutil
import os
from pathlib import Path

from geniac.cli.commands.init import GeniacInit

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacConfigs(GeniacInit):
    """Geniac CMake Configs class"""

    def __init__(
        self,
        *args,
        src_path: str = None,
        **kwargs,
    ):
        """Init flags specific to GConfigs command"""
        super().__init__(
            *args, src_path=src_path, post_clean=True, init_build=False, **kwargs
        )

    def cmake_configs(self):
        """Generate config files"""

        cmake_run = (
            self._subprocess_run(
                ["cmake", "-DCMAKE_INSTALL_PREFIX=" + (self.working_dirs["src"]).as_posix() + "/../.install", (self.working_dirs["src"] / "geniac").as_posix()],
                check=True,
                capture_output=True,
                cwd=self.working_dirs["build"],
                cmd_name="CMake run",
            )
            if self.working_dirs.get("src")
            else None
        )

        make_run = (
            self._subprocess_run(
                ["make", "build_config_files"],
                capture_output=True,
                check=True,
                cwd=self.working_dirs["build"],
                cmd_name="Make config files",
            )
            if self.working_dirs.get("src")
            else None
        )
        output_dir = Path(self.working_dirs["build"].as_posix() + "/../configs").as_posix()
        output_dir = os.path.realpath(output_dir)
        if os.path.isdir(output_dir):
            shutil.rmtree(output_dir)
            self.info("The folder '%s' has been deleted and will be reinitialized.", output_dir)

        shutil.copytree((self.working_dirs["build"]).as_posix() + "/workDir/results/conf/", output_dir, ignore=shutil.ignore_patterns("*.txt"))
        self.info("The '%s' folder has been created with the config files generated by geniac.", output_dir)

    def run(self):
        """

        Returns:

        """
        self.cmake_configs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""options.py: Geniac CMake option class"""

import logging

from geniac.cli.commands.install import GeniacInstall

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"

_logger = logging.getLogger(__name__)


class GeniacTest(GeniacInstall):
    """Geniac CMake option class"""

    def __init__(
        self,
        profile: list,
        *args,
        src_path: str = None,
        check_cluster: bool = False,
        **kwargs,
    ):
        """Init flags specific to GOptions command"""
        self.check_cluster = check_cluster
        self.nxf_profiles = profile
        # Give the profile asked as mode opt in install cmd
        super().__init__(
            *args,
            src_path=src_path,
            mode=profile,
            config_section="install",
            **kwargs,
        )

    def test(self):
        """Launch Make test commands in the folder"""
        for nxf_profile in self.nxf_profiles:
            self._subprocess_run(
                [
                    "make",
                    f"test_{nxf_profile + '_cluster' if self.check_cluster else nxf_profile}",
                ],
                capture_output=True,
                cmd_name="Make Test",
                check=True,
                cwd=self.working_dirs["build"],
            )

    def run(self):
        """

        Returns:

        """
        #self.build()
        self.test()
